from typing import Dict, Union, Type, Optional
from lark import v_args
from elasticsearch_dsl import Q, Text, Keyword, Integer, Field

from optimade.filtertransformers import BaseTransformer, Quantity
from optimade.server.mappers import BaseResourceMapper

__all__ = ("ElasticTransformer",)


class ElasticsearchQuantity(Quantity):
    """Elasticsearch-specific extension of the underlying
    [`Quantity`][optimade.filtertransformers.base_transformer.Quantity] class.

    Attributes:
        name: The name of the quantity as used in the filter expressions.
        backend_field: The name of the field for this quantity in Elasticsearch, will be
            ``name`` by default.
        elastic_mapping_type: A decendent of an `elasticsearch_dsl.Field` that denotes which
            mapping type was used in the Elasticsearch index.
        length_quantity: Elasticsearch does not support length of arrays, but we can
            map fields with array to other fields with ints about the array length. The
            LENGTH operator will only be supported for quantities with this attribute.
        has_only_quantity: Elasticsearch does not support exclusive search on arrays, like
            a list of chemical elements. But, we can order all elements by atomic number
            and use a keyword field with all elements to perform this search. This only
            works for elements (i.e. labels in ``CHEMICAL_SYMBOLS``) and quantities
            with this attribute.
        nested_quantity: To support optimade's 'zipped tuple' feature (e.g.
            'elements:elements_ratios HAS "H":>0.33), we use elasticsearch nested objects
            and nested queries. This quantity will provide the field for the nested
            object that contains the quantity (and others). The zipped tuples will only
            work for quantities that share the same nested object quantity.
    """

    name: str
    backend_field: Optional[str]
    length_quantity: Optional["ElasticsearchQuantity"]
    elastic_mapping_type: Optional[Field]
    has_only_quantity: Optional["ElasticsearchQuantity"]
    nested_quantity: Optional["ElasticsearchQuantity"]

    def __init__(
        self,
        name: str,
        backend_field: str = None,
        length_quantity: "ElasticsearchQuantity" = None,
        elastic_mapping_type: Field = None,
        has_only_quantity: "ElasticsearchQuantity" = None,
        nested_quantity: "ElasticsearchQuantity" = None,
    ):
        """Initialise the quantity from its name, aliases and mapping type.

        Parameters:
            name: The name of the quantity as used in the filter expressions.
            backend_field: The name of the field for this quantity in Elasticsearch, will be
                ``name`` by default.
            elastic_mapping_type: A decendent of an `elasticsearch_dsl.Field` that denotes which
                mapping type was used in the Elasticsearch index.
            length_quantity: Elasticsearch does not support length of arrays, but we can
                map fields with array to other fields with ints about the array length. The
                LENGTH operator will only be supported for quantities with this attribute.
            has_only_quantity: Elasticsearch does not support exclusive search on arrays, like
                a list of chemical elements. But, we can order all elements by atomic number
                and use a keyword field with all elements to perform this search. This only
                works for elements (i.e. labels in ``CHEMICAL_SYMBOLS``) and quantities
                with this attribute.
            nested_quantity: To support optimade's 'zipped tuple' feature (e.g.
                'elements:elements_ratios HAS "H":>0.33), we use elasticsearch nested objects
                and nested queries. This quantity will provide the field for the nested
                object that contains the quantity (and others). The zipped tuples will only
                work for quantities that share the same nested object quantity.
        """

        super().__init__(name, backend_field, length_quantity)

        self.elastic_mapping_type = (
            Keyword if elastic_mapping_type is None else elastic_mapping_type
        )
        self.has_only_quantity = has_only_quantity
        self.nested_quantity = nested_quantity


class ElasticTransformer(BaseTransformer):
    """Transformer that transforms ``v0.10.1``/`v1.0` grammar parse
    trees into Elasticsearch queries.

    Uses elasticsearch_dsl and will produce an `elasticsearch_dsl.Q` instance.

    """

    operator_map = {
        "<": "lt",
        "<=": "lte",
        ">": "gt",
        ">=": "gte",
    }

    _quantity_type: Type[ElasticsearchQuantity] = ElasticsearchQuantity

    def __init__(
        self, mapper: BaseResourceMapper = None, quantities: Dict[str, Quantity] = None
    ):
        if quantities is not None:
            self.quantities = quantities

        super().__init__(mapper=mapper)

    def _field(self, quantity: Union[str, Quantity], nested: Quantity = None) -> str:
        """Used to unwrap from `property` to the string backend field name.

        If passed a `Quantity` (or a derived `ElasticsearchQuantity`), this method
        returns the backend field name, modulo some handling of nested fields.

        If passed a string quantity name:
        - Check that the name does not match a relationship type,
          raising a `NotImplementedError` if it does.
        - If the string is prefixed by an underscore, assume this is a
          provider-specific field from another provider and simply return it.
          The original `property` rule would have already filtered out provider
          fields for this backend appropriately as `Quantity` objects.

        Returns:
            The field name to use for database queries.

        """

        if isinstance(quantity, str):
            if quantity in self.mapper.RELATIONSHIP_ENTRY_TYPES:
                raise NotImplementedError(
                    f"Unable to filter on relationships with type {quantity!r}"
                )

            # In this case, the property rule has already filtered out fields
            # that do not match this provider, so this indicates an "other provider"
            # field that should be passed over
            if quantity.startswith("_"):
                return quantity

        if nested is not None:
            return "%s.%s" % (nested.backend_field, quantity.name)

        return quantity.backend_field

    def _query_op(
        self,
        quantity: Union[ElasticsearchQuantity, str],
        op: str,
        value: Union[str, float, int],
        nested: ElasticsearchQuantity = None,
    ) -> Q:
        """Return a range, match, or term query for the given quantity, comparison
        operator, and value.

        Returns:
            An elasticsearch_dsl query.

        Raises:
            BadRequest: If the query is not well-defined or is not supported.
        """
        field = self._field(quantity, nested=nested)
        if op in self.operator_map:
            return Q("range", **{field: {self.operator_map[op]: value}})

        # If quantity is an "other provider" field then use Keyword as the default
        # mapping type. These queries should not match on anything as the field
        # is not present in the index.
        elastic_mapping_type = Keyword
        if isinstance(quantity, ElasticsearchQuantity):
            elastic_mapping_type = quantity.elastic_mapping_type

        if elastic_mapping_type == Text:
            query_type = "match"
        elif elastic_mapping_type in [Keyword, Integer]:
            query_type = "term"
        else:
            raise NotImplementedError("Quantity has unsupported ES field type")

        if op in ["=", ""]:
            return Q(query_type, **{field: value})

        if op == "!=":
            # != queries must also include an existence check
            # Note that for MongoDB, `$exists` will include null-valued fields,
            # where as in ES `exists` excludes them.
            # pylint: disable=invalid-unary-operand-type
            return ~Q(query_type, **{field: value}) & Q("exists", field=field)

    def _has_query_op(self, quantities, op, predicate_zip_list):
        """Returns a bool query that combines the operator calls `_query_op`
        for each predicate and zipped quantity predicate combination.
        """
        if op == "HAS":
            kind = "must"  # in case of HAS we do a must over the "list" of the one given element
        elif op == "HAS ALL":
            kind = "must"
        elif op == "HAS ANY":
            kind = "should"
        elif op == "HAS ONLY":
            # HAS ONLY comes with heavy limitations, because there is no such thing
            # in elastic search. Only supported for elements, where we can construct
            # an anonymous "formula" based on elements sorted by order number and
            # can do a = comparision to check if all elements are contained

            # @ml-evs: Disabling this HAS ONLY workaround as tests are not passing
            raise NotImplementedError(
                "HAS ONLY queries are not currently supported by the Elasticsearch backend."
            )

            # from optimade.models import CHEMICAL_SYMBOLS, ATOMIC_NUMBERS

            # if len(quantities) > 1:
            #     raise NotImplementedError("HAS ONLY is not supported with zip")
            # quantity = quantities[0]

            # if quantity.has_only_quantity is None:
            #     raise NotImplementedError(
            #         "HAS ONLY is not supported by %s" % quantity.name
            #     )

            # def values():
            #     for predicates in predicate_zip_list:
            #         if len(predicates) != 1:
            #             raise NotImplementedError("Tuples not supported in HAS ONLY")
            #         op, value = predicates[0]
            #         if op != "=":
            #             raise NotImplementedError(
            #                 "Predicated not supported in HAS ONLY"
            #             )
            #         if not isinstance(value, str):
            #             raise NotImplementedError("Only strings supported in HAS ONLY")
            #         yield value

            # try:
            #     order_numbers = list([ATOMIC_NUMBERS[element] for element in values()])
            #     order_numbers.sort()
            #     value = "".join(
            #         [CHEMICAL_SYMBOLS[number - 1] for number in order_numbers]
            #     )
            # except KeyError:
            #     raise NotImplementedError(
            #         "HAS ONLY is only supported for chemical symbols"
            #     )

            # return Q("term", **{quantity.has_only_quantity.name: value})
        else:
            raise NotImplementedError(f"Unrecognised operation {op}.")

        queries = [
            self._has_query(quantities, predicates) for predicates in predicate_zip_list
        ]
        return Q("bool", **{kind: queries})

    def _has_query(self, quantities, predicates):
        """
        Returns a bool query that combines the operator queries ():func:`_query_op`)
        for quantity pericate combination.
        """
        if len(quantities) != len(predicates):
            raise ValueError(
                "Tuple length does not match: %s <o> %s "
                % (":".join(quantities), ":".join(predicates))
            )

        if len(quantities) == 1:
            o, value = predicates[0]
            return self._query_op(quantities[0], o, value)

        nested_quantity = quantities[0].nested_quantity
        same_nested_quantity = any(
            q.nested_quantity != nested_quantity for q in quantities
        )
        if nested_quantity is None or same_nested_quantity:
            raise NotImplementedError(
                "Expression with tuples are only supported for %s"
                % ", ".join(quantities)
            )

        queries = [
            self._query_op(quantity, o, value, nested=nested_quantity)
            for quantity, (o, value) in zip(quantities, predicates)
        ]

        return Q(
            "nested",
            path=self._field(nested_quantity),
            query=dict(bool=dict(must=queries)),
        )

    def __default__(self, tree, children, *args, **kwargs):
        """Default behavior for rules that only replace one symbol with another"""
        return children[0]

    def filter(self, args):
        # filter: expression*
        if len(args) == 1:
            return args[0]
        return Q("bool", **{"must": args})

    def expression_clause(self, args):
        # expression_clause: expression_phrase ( _AND expression_phrase )*
        result = args[0]
        for arg in args[1:]:
            result &= arg
        return result

    def expression(self, args):
        # expression: expression_clause ( _OR expression_clause )*
        result = args[0]
        for arg in args[1:]:
            result |= arg
        return result

    def expression_phrase(self, args):
        # expression_phrase: [ NOT ] ( operator | "(" expression ")" )
        if args[0] == "NOT":
            return ~args[1]
        return args[0]

    @v_args(inline=True)
    def property_first_comparison(self, quantity, query):
        # property_first_comparison: property *_rhs
        return query(quantity)

    @v_args(inline=True)
    def constant_first_comparison(self, value, op, quantity):
        # constant_first_comparison: constant OPERATOR ( non_string_value | ...not_implemented_string )
        if not isinstance(quantity, Quantity):
            raise TypeError("Only quantities can be compared to constant values.")

        return self._query_op(quantity, self._reversed_operator_map[op], value)

    @v_args(inline=True)
    def value_op_rhs(self, op, value):
        # value_op_rhs: OPERATOR value
        return lambda quantity: self._query_op(quantity, op, value)

    def length_op_rhs(self, args):
        # length_op_rhs: LENGTH [ OPERATOR ] signed_int
        value = args[-1]
        if len(args) == 3:
            op = args[1]
        else:
            op = "="

        def query(quantity):

            # This is only the case if quantity is an "other" provider's field,
            # in which case, we should treat it as unknown and try to do a null query
            if isinstance(quantity, str):
                return self._query_op(quantity, op, value)

            if quantity.length_quantity is None:
                raise NotImplementedError(
                    f"LENGTH is not supported for {quantity.name!r}"
                )
            quantity = quantity.length_quantity
            return self._query_op(quantity, op, value)

        return query

    @v_args(inline=True)
    def known_op_rhs(self, _, value):
        # known_op_rhs: IS ( KNOWN | UNKNOWN )

        def query(quantity):
            query = Q("exists", field=self._field(quantity))
            if value == "KNOWN":
                return query
            elif value == "UNKNOWN":
                return ~query  # pylint: disable=invalid-unary-operand-type
            raise NotImplementedError

        return query

    def set_op_rhs(self, args):
        # set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ... )
        values = args[-1]
        if not isinstance(values, list):
            if len(args) == 3:
                op = args[1]
            else:
                op = "="
            values = [(op, values)]

        if len(args) == 3:
            op = "HAS " + args[1]
        else:
            op = "HAS"

        return lambda quantity: self._has_query_op(
            [quantity], op, [[value] for value in values]
        )

    def set_zip_op_rhs(self, args):
        # set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list | ANY value_zip_list )
        add_on = args[0]
        values = args[-1]
        if len(args) == 4:
            op = "HAS " + args[2]
        else:
            op = "HAS"
            values = [values]

        return lambda quantity: self._has_query_op([quantity] + add_on, op, values)

    def property_zip_addon(self, args):
        raise NotImplementedError("Correlated list queries are not supported.")
        return args

    def value_zip(self, args):
        raise NotImplementedError("Correlated list queries are not supported.")
        return self.value_list(args)

    def value_zip_list(self, args):
        raise NotImplementedError("Correlated list queries are not supported.")
        return args

    def value_list(self, args):
        result = []
        op = "="
        for arg in args:
            if arg in ["<", "<=", ">", ">=", "!=", "="]:
                op = arg
            else:
                result.append(
                    (
                        op,
                        arg,
                    )
                )
                op = "="
        return result

    def fuzzy_string_op_rhs(self, args):
        op = args[0]
        value = args[-1]
        if op == "CONTAINS":
            wildcard = "*%s*" % value
        if op == "STARTS":
            wildcard = "%s*" % value
        if op == "ENDS":
            wildcard = "*%s" % value

        return lambda quantity: Q("wildcard", **{self._field(quantity): wildcard})

    @v_args(inline=True)
    def string(self, string):
        # string: ESCAPED_STRING
        return string.strip('"')

    @v_args(inline=True)
    def signed_int(self, number):
        # signed_int : SIGNED_INT
        return int(number)

    @v_args(inline=True)
    def number(self, number):
        # number: SIGNED_INT | SIGNED_FLOAT
        if number.type == "SIGNED_INT":
            type_ = int
        elif number.type == "SIGNED_FLOAT":
            type_ = float
        return type_(number)



# pylint: disable=line-too-long,no-self-argument
from pydantic import (  # pylint: disable=no-name-in-module
    BaseModel,
    AnyUrl,
    validator,
)
from typing import List, Optional

from optimade.models.entries import EntryResource, EntryResourceAttributes
from optimade.models.utils import OptimadeField, SupportLevel


__all__ = ("Person", "ReferenceResourceAttributes", "ReferenceResource")


class Person(BaseModel):
    """A person, i.e., an author, editor or other."""

    name: str = OptimadeField(
        ...,
        description="""Full name of the person, REQUIRED.""",
        support=SupportLevel.MUST,
        queryable=SupportLevel.OPTIONAL,
    )

    firstname: Optional[str] = OptimadeField(
        None,
        description="""First name of the person.""",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    lastname: Optional[str] = OptimadeField(
        None,
        description="""Last name of the person.""",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )


class ReferenceResourceAttributes(EntryResourceAttributes):
    """Model that stores the attributes of a reference.

    Many properties match the meaning described in the
    [BibTeX specification](http://bibtexml.sourceforge.net/btxdoc.pdf).

    """

    authors: Optional[List[Person]] = OptimadeField(
        None,
        description="List of person objects containing the authors of the reference.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    editors: Optional[List[Person]] = OptimadeField(
        None,
        description="List of person objects containing the editors of the reference.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    doi: Optional[str] = OptimadeField(
        None,
        description="The digital object identifier of the reference.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    url: Optional[AnyUrl] = OptimadeField(
        None,
        description="The URL of the reference.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    address: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    annote: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    booktitle: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    chapter: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    crossref: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    edition: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    howpublished: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    institution: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    journal: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    key: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    month: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    note: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    number: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    organization: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    pages: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    publisher: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    school: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    series: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    title: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    bib_type: Optional[str] = OptimadeField(
        None,
        description="Type of the reference, corresponding to the **type** property in the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    volume: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )

    year: Optional[str] = OptimadeField(
        None,
        description="Meaning of property matches the BiBTeX specification.",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.OPTIONAL,
    )


class ReferenceResource(EntryResource):
    """The `references` entries describe bibliographic references.

    The following properties are used to provide the bibliographic details:

    - **address**, **annote**, **booktitle**, **chapter**, **crossref**, **edition**, **howpublished**, **institution**, **journal**, **key**, **month**, **note**, **number**, **organization**, **pages**, **publisher**, **school**, **series**, **title**, **volume**, **year**: meanings of these properties match the [BibTeX specification](http://bibtexml.sourceforge.net/btxdoc.pdf), values are strings;
    - **bib_type**: type of the reference, corresponding to **type** property in the BibTeX specification, value is string;
    - **authors** and **editors**: lists of *person objects* which are dictionaries with the following keys:
        - **name**: Full name of the person, REQUIRED.
        - **firstname**, **lastname**: Parts of the person's name, OPTIONAL.
    - **doi** and **url**: values are strings.
    - **Requirements/Conventions**:
        - **Support**: OPTIONAL support in implementations, i.e., any of the properties MAY be `null`.
        - **Query**: Support for queries on any of these properties is OPTIONAL.
            If supported, filters MAY support only a subset of comparison operators.
        - Every references entry MUST contain at least one of the properties.

    """

    type: str = OptimadeField(
        "references",
        const="references",
        description="""The name of the type of an entry.
- **Type**: string.
- **Requirements/Conventions**:
    - **Support**: MUST be supported by all implementations, MUST NOT be `null`.
    - **Query**: MUST be a queryable property with support for all mandatory filter features.
    - **Response**: REQUIRED in the response.
    - MUST be an existing entry type.
    - The entry of type <type> and ID <id> MUST be returned in response to a request for `/<type>/<id>` under the versioned base URL.
- **Example**: `"structures"`""",
        pattern="^references$",
        support=SupportLevel.MUST,
        queryable=SupportLevel.MUST,
    )
    attributes: ReferenceResourceAttributes

    @validator("attributes")
    def validate_attributes(cls, v):
        if not any(prop[1] is not None for prop in v):
            raise ValueError("reference object must have at least one field defined")
        return v



"""This module contains the
[`ImplementationValidator`][optimade.validator.validator.ImplementationValidator]
class that can be pointed at an OPTIMADE implementation and validated
against the specification via the pydantic models implemented in this package.

"""
# pylint: disable=import-outside-toplevel

import re
import sys
import logging
import random
import urllib.parse
import dataclasses
from typing import Union, Tuple, Any, List, Dict, Optional

try:
    import simplejson as json
except ImportError:
    import json

import requests

from optimade.models import DataType, EntryInfoResponse, SupportLevel
from optimade.validator.utils import (
    Client,
    test_case,
    print_failure,
    print_notify,
    print_success,
    print_warning,
    ResponseError,
    ValidatorEntryResponseOne,
    ValidatorEntryResponseMany,
    ValidatorResults,
)

from optimade.validator.config import VALIDATOR_CONFIG as CONF

VERSIONS_REGEXP = r".*/v[0-9]+(\.[0-9]+){,2}$"

__all__ = ("ImplementationValidator",)


class ImplementationValidator:
    """Class used to make a series of checks against a particular
    OPTIMADE implementation over HTTP.

    Uses the pydantic models in [`optimade.models`][optimade.models] to
    validate the response from the server and crawl through the
    available endpoints.

    Attributes:
        valid: whether or not the implementation was deemed valid, with
            `None` signifying that tests did not run.

    Caution:
        Only works for current version of the specification as defined
        by [`optimade.models`][optimade.models].

    """

    valid: Optional[bool]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        client: Any = None,
        base_url: str = None,
        verbosity: int = 0,
        respond_json: bool = False,
        page_limit: int = 5,
        max_retries: int = 5,
        run_optional_tests: bool = True,
        fail_fast: bool = False,
        as_type: str = None,
        index: bool = False,
        minimal: bool = False,
        http_headers: Dict[str, str] = None,
    ):
        """Set up the tests to run, based on constants in this module
        for required endpoints.

        Arguments:
            client: A client that has a `.get()` method to obtain the
                response from the implementation. If `None`, then
                [`Client`][optimade.validator.utils.Client] will be used.
            base_url: The URL of the implementation to validate. Unless
                performing "as_type" validation, this should point to the
                base of the OPTIMADE implementation.
            verbosity: The verbosity of the output and logging as an integer
                (`0`: critical, `1`: warning, `2`: info, `3`: debug).
            respond_json: If `True`, print only a JSON representation of the
                results of validation to stdout.
            page_limit: The default page limit to apply to filters.
            max_retries: Argument is passed to the client for how many
                attempts to make for a request before failing.
            run_optional_tests: Whether to run the tests on optional
                OPTIMADE features.
            fail_fast: Whether to exit validation after the first failure
                of a mandatory test.
            as_type: An OPTIMADE entry or endpoint type to coerce the response
                from implementation into, e.g. "structures". Requires `base_url`
                to be pointed to the corresponding endpoint.
            index: Whether to validate the implementation as an index meta-database.
            minimal: Whether or not to run only a minimal test set.
            http_headers: Dictionary of additional headers to add to every request.

        """
        self.verbosity = verbosity
        self.max_retries = max_retries
        self.page_limit = page_limit
        self.index = index
        self.run_optional_tests = run_optional_tests
        self.fail_fast = fail_fast
        self.respond_json = respond_json
        self.minimal = minimal

        if as_type is None:
            self.as_type_cls = None
        elif self.index:
            if as_type not in CONF.response_classes_index:
                raise RuntimeError(
                    f"Provided as_type='{as_type}' not allowed for an Index meta-database."
                )
            self.as_type_cls = CONF.response_classes_index[as_type]
        elif as_type in ("structure", "reference"):
            self.as_type_cls = CONF.response_classes[f"{as_type}s/"]
        else:
            self.as_type_cls = CONF.response_classes[as_type]

        if client is None and base_url is None:
            raise RuntimeError(
                "Need at least a URL or a client to initialize validator."
            )
        if base_url and client:
            raise RuntimeError("Please specify at most one of base_url or client.")
        if client:
            self.client = client
            self.base_url = self.client.base_url
            # If a custom client has been provided, try to set custom headers if they have been specified,
            # but do not overwrite any existing attributes
            if http_headers:
                if not hasattr(self.client, "headers"):
                    self.client.headers = http_headers
                else:
                    print_warning(
                        f"Not using specified request headers {http_headers} with custom client {self.client}."
                    )
        else:
            while base_url.endswith("/"):
                base_url = base_url[:-1]
            self.base_url = base_url
            self.client = Client(
                base_url, max_retries=self.max_retries, headers=http_headers
            )

        self._setup_log()

        self._response_classes = (
            CONF.response_classes_index if self.index else CONF.response_classes
        )

        # some simple checks on base_url
        self.base_url_parsed = urllib.parse.urlparse(self.base_url)
        # only allow filters/endpoints if we are working in "as_type" mode
        if self.as_type_cls is None and self.base_url_parsed.query:
            raise SystemExit("Base URL not appropriate: should not contain a filter.")

        self.valid = None

        self._test_id_by_type = {}
        self._entry_info_by_type = {}

        self.results = ValidatorResults(verbosity=self.verbosity)

    def _setup_log(self):
        """Define stdout log based on given verbosity."""
        self._log = logging.getLogger("optimade").getChild("validator")
        self._log.handlers = []
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s | %(levelname)8s: %(message)s")
        )

        if not self.respond_json:
            self._log.addHandler(stdout_handler)
        else:
            self.verbosity = -1

        if self.verbosity == 0:
            self._log.setLevel(logging.CRITICAL)
        elif self.verbosity == 1:
            self._log.setLevel(logging.WARNING)
        elif self.verbosity == 2:
            self._log.setLevel(logging.INFO)
        elif self.verbosity > 0:
            self._log.setLevel(logging.DEBUG)

    def print_summary(self):
        """Print a summary of the results of validation."""
        if self.respond_json:
            print(json.dumps(dataclasses.asdict(self.results), indent=2))
            return

        if self.results.failure_messages:
            print("\n\nFAILURES")
            print("========\n")
            for message in self.results.failure_messages:
                print_failure(message[0])
                for line in message[1].split("\n"):
                    print_warning("\t" + line)

        if self.results.optional_failure_messages:
            print("\n\nOPTIONAL TEST FAILURES")
            print("======================\n")
            for message in self.results.optional_failure_messages:
                print_notify(message[0])
                for line in message[1].split("\n"):
                    print_warning("\t" + line)

        if self.results.internal_failure_messages:
            print("\n\nINTERNAL FAILURES")
            print("=================\n")
            print(
                "There were internal validator failures associated with this run.\n"
                "If this problem persists, please report it at:\n"
                "https://github.com/Materials-Consortia/optimade-python-tools/issues/new\n"
            )

            for message in self.results.internal_failure_messages:
                print_warning(message[0])
                for line in message[1].split("\n"):
                    print_warning("\t" + line)

        if self.valid or (not self.valid and not self.fail_fast):
            final_message = f"\n\nPassed {self.results.success_count} out of {self.results.success_count + self.results.failure_count + self.results.internal_failure_count} tests."
            if not self.valid:
                print_failure(final_message)
            else:
                print_success(final_message)

            if self.run_optional_tests and not self.fail_fast:
                print(
                    f"Additionally passed {self.results.optional_success_count} out of "
                    f"{self.results.optional_success_count + self.results.optional_failure_count} optional tests."
                )

    def validate_implementation(self):
        """Run all the test cases on the implementation, or the single type test,
        depending on what options were provided on initialiation.

        Sets the `self.valid` attribute to `True` or `False` depending on the
        outcome of the tests.

        Raises:
            RuntimeError: If it was not possible to start the validation process.

        """
        # If a single "as type" has been set, only run that test
        if self.as_type_cls is not None:
            self._log.debug(
                "Validating response of %s with model %s",
                self.base_url,
                self.as_type_cls,
            )
            self._test_as_type()
            self.valid = not bool(self.results.failure_count)
            self.print_summary()
            return

        # Test entire implementation
        if self.verbosity >= 0:
            print(f"Testing entire implementation at {self.base_url}")
        info_endp = CONF.info_endpoint
        self._log.debug("Testing base info endpoint of %s", info_endp)

        # Get and validate base info to find endpoints
        # If this is not possible, then exit at this stage
        base_info = self._test_info_or_links_endpoint(info_endp)
        if not base_info:
            self._log.critical(
                f"Unable to deserialize response from introspective {info_endp!r} endpoint. "
                "This is required for all further validation, so the validator will now exit."
            )
            # Set valid to False to ensure error code 1 is raised at CLI
            self.valid = False
            self.print_summary()
            return

        # Grab the provider prefix from base info and use it when looking for provider fields
        self.provider_prefix = None
        meta = base_info.get("meta", {})
        if meta.get("provider") is not None:
            self.provider_prefix = meta["provider"].get("prefix")

        # Set the response class for all `/info/entry` endpoints based on `/info` response
        self.available_json_endpoints, _ = self._get_available_endpoints(
            base_info, request=info_endp
        )
        for endp in self.available_json_endpoints:
            self._response_classes[f"{info_endp}/{endp}"] = EntryInfoResponse

        # Run some tests on the versions endpoint
        self._log.debug("Testing versions endpoint %s", CONF.versions_endpoint)
        self._test_versions_endpoint()
        self._test_bad_version_returns_553()

        # Test that entry info endpoints deserialize correctly
        # If they do not, the corresponding entry in _entry_info_by_type
        # is set to False, which must be checked for further validation
        for endp in self.available_json_endpoints:
            entry_info_endpoint = f"{info_endp}/{endp}"
            self._log.debug("Testing expected info endpoint %s", entry_info_endpoint)
            self._entry_info_by_type[endp] = self._test_info_or_links_endpoint(
                entry_info_endpoint
            )

        # Use the _entry_info_by_type to construct filters on the relevant endpoints
        if not self.minimal:
            for endp in self.available_json_endpoints:
                self._log.debug("Testing queries on JSON entry endpoint of %s", endp)
                self._recurse_through_endpoint(endp)

        # Test that the results from multi-entry-endpoints obey, e.g. page limits,
        # and that all entries can be deserialized with the patched models.
        # These methods also set the test_ids for each type of entry, which are validated
        # in the next loop.
        for endp in self.available_json_endpoints:
            self._log.debug("Testing multiple entry endpoint of %s", endp)
            self._test_multi_entry_endpoint(endp)

        # Test that the single IDs scraped earlier work with the single entry endpoint
        for endp in self.available_json_endpoints:
            self._log.debug("Testing single entry request of type %s", endp)
            self._test_single_entry_endpoint(endp)

        # Test that the links endpoint can be serialized correctly
        self._log.debug("Testing %s endpoint", CONF.links_endpoint)
        self._test_info_or_links_endpoint(CONF.links_endpoint)

        self.valid = not (
            self.results.failure_count or self.results.internal_failure_count
        )

        self.print_summary()

    @test_case
    def _recurse_through_endpoint(self, endp: str) -> Tuple[bool, str]:
        """For a given endpoint (`endp`), get the entry type
        and supported fields, testing that all mandatory fields
        are supported, then test queries on every property according
        to the reported type, with optionality decided by the
        specification-level support level for that field.

        Parameters:
            endp: Endpoint to be tested.

        Returns:
            `True` if endpoint passed the tests, and a string summary.

        """
        entry_info = self._entry_info_by_type.get(endp)

        if not entry_info:
            raise ResponseError(
                f"Unable to generate filters for endpoint {endp}: 'info/{endp}' response was malformed."
            )

        _impl_properties = self._check_entry_info(entry_info, endp)
        prop_list = list(_impl_properties.keys())

        self._check_response_fields(endp, prop_list)
        chosen_entry, _ = self._get_archetypal_entry(endp, prop_list)

        if not chosen_entry:
            return (
                None,
                f"Unable to generate filters for endpoint {endp}: no valid entries found.",
            )

        for prop in _impl_properties:
            # check support level of property
            prop_type = _impl_properties[prop]["type"]
            sortable = _impl_properties[prop]["sortable"]
            optional = (
                CONF.entry_schemas[endp].get(prop, {}).get("queryable")
                == SupportLevel.OPTIONAL
            )

            if optional and not self.run_optional_tests:
                continue

            self._construct_queries_for_property(
                prop,
                prop_type,
                sortable,
                endp,
                chosen_entry,
                request=f"; testing queries for {endp}->{prop}",
                optional=optional,
            )

        self._test_unknown_provider_property(endp)
        self._test_completely_unknown_property(endp)

        return True, f"successfully recursed through endpoint {endp}."

    @test_case
    def _test_completely_unknown_property(self, endp):
        request = f"{endp}?filter=crazyfield = 2"
        response, _ = self._get_endpoint(
            request,
            expected_status_code=400,
        )

        return True, "unknown field returned 400 Bad Request, as expected"

    @test_case
    def _test_unknown_provider_property(self, endp):

        dummy_provider_field = "_crazyprovider_field"

        request = f"{endp}?filter={dummy_provider_field}=2"
        response, _ = self._get_endpoint(
            request,
            multistage=True,
            request=request,
        )

        if response is not None:
            deserialized, _ = self._deserialize_response(
                response, CONF.response_classes[endp], request=request, multistage=True
            )

            return (
                True,
                "Unknown provider field was ignored when filtering, as expected",
            )

        raise ResponseError(
            "Failed to handle field from unknown provider; should return without affecting filter results"
        )

    def _check_entry_info(self, entry_info: Dict[str, Any], endp: str) -> List[str]:
        """Checks that `entry_info` contains all the required properties,
        and returns the property list for the endpoint.

        Parameters:
            entry_info: JSON representation of the response from the
                entry info endpoint.
            endp: The name of the entry endpoint.

        Returns:
            The list of property names supported by this implementation.

        """
        properties = entry_info.get("data", {}).get("properties", [])
        self._test_must_properties(
            properties, endp, request=f"{CONF.info_endpoint}/{endp}"
        )

        return properties

    @test_case
    def _test_must_properties(
        self, properties: List[str], endp: str
    ) -> Tuple[bool, str]:
        """Check that the entry info lists all properties with the "MUST"
        support level for this endpoint.

        Parameters:
            properties: The list of property names supported by the endpoint.
            endp: The endpoint.

        Returns:
            `True` if the properties were found, and a string summary.

        """
        must_props = set(
            prop
            for prop in CONF.entry_schemas.get(endp, {})
            if CONF.entry_schemas[endp].get(prop, {}).get("support")
            == SupportLevel.MUST
        )
        must_props_supported = set(prop for prop in properties if prop in must_props)
        missing = must_props - must_props_supported
        if len(missing) != 0:
            raise ResponseError(
                f"Some 'MUST' properties were missing from info/{endp}: {missing}"
            )

        return True, f"Found all required properties in entry info for endpoint {endp}"

    @test_case
    def _get_archetypal_entry(
        self, endp: str, properties: List[str]
    ) -> Tuple[Dict[str, Any], str]:
        """Get a random entry from the first page of results for this
        endpoint.

        Parameters:
            endp: The endpoint to query.

        Returns:
            The JSON representation of the chosen entry and the summary message.


        """
        response, message = self._get_endpoint(endp, multistage=True)

        if response:
            data = response.json().get("data", [])
            data_returned = len(data)
            if data_returned < 1:
                return (
                    None,
                    "Endpoint {endp!r} returned no entries, cannot get archetypal entry.",
                )

            response, message = self._get_endpoint(
                f"{endp}?page_offset={random.randint(0, data_returned-1)}&response_fields={','.join(properties)}",
                multistage=True,
            )
            if response:
                archetypal_entry = response.json()["data"][0]
                return (
                    archetypal_entry,
                    f"set archetypal entry for {endp} with ID {archetypal_entry['id']}.",
                )

        raise ResponseError(f"Failed to get archetypal entry. Details: {message}")

    @test_case
    def _check_response_fields(self, endp: str, fields: List[str]) -> Tuple[bool, str]:
        """Check that the response field query parameter is obeyed.

        Parameters:
            endp: The endpoint to query.
            fields: The known fields for this endpoint to test.

        Returns:
            Bool indicating success and a summary message.

        """

        subset_fields = random.sample(fields, min(len(fields) - 1, 3))
        test_query = f"{endp}?response_fields={','.join(subset_fields)}&page_limit=1"
        response, _ = self._get_endpoint(test_query, multistage=True)

        if response and len(response.json()["data"]) == 1:
            doc = response.json()["data"][0]
            expected_fields = set(subset_fields)
            expected_fields -= CONF.top_level_non_attribute_fields

            if "attributes" not in doc:
                raise ResponseError(
                    f"Entries are missing `attributes` key.\nReceived: {doc}"
                )

            returned_fields = set(sorted(list(doc.get("attributes", {}).keys())))
            returned_fields -= CONF.top_level_non_attribute_fields

            if expected_fields != returned_fields:
                raise ResponseError(
                    f"Response fields not obeyed by {endp!r}:\nExpected: {expected_fields}\nReturned: {returned_fields}"
                )

            return True, "Successfully limited response fields"

        return (
            None,
            f"Unable to test adherence to response fields as no entries were returned for endpoint {endp!r}.",
        )

    @test_case
    def _construct_queries_for_property(
        self,
        prop: str,
        prop_type: DataType,
        sortable: bool,
        endp: str,
        chosen_entry: Dict[str, Any],
    ) -> Tuple[Optional[bool], str]:
        """For the given property, property type and chose entry, this method
        runs a series of queries for each field in the entry, testing that the
        initial document is returned where expected.


        Parameters:
            prop: The property name.
            prop_type: The property type.
            sortable: Whether the implementation has indicated that the field is sortable.
            endp: The corresponding entry endpoint.
            chosen_entry: A JSON respresentation of the chosen entry that will be used to
                construct the filters.

        Returns:
            Boolean indicating success (`True`) or failure/irrelevance
            (`None`) and the string summary of the test case.

        """
        # Explicitly handle top level keys that do not have types in info
        if not chosen_entry:
            raise ResponseError(
                f"Chosen entry of endpoint '/{endp}' failed validation."
            )

        if prop == "type":
            if chosen_entry["type"] == endp:
                return True, f"Successfully validated {prop}"
            raise ResponseError(
                f"Chosen entry of endpoint '{endp}' had unexpected type {chosen_entry['type']!r}."
            )

        prop_type = (
            CONF.entry_schemas[endp].get(prop, {}).get("type")
            if prop_type is None
            else prop_type
        )

        if prop_type is None:
            raise ResponseError(
                f"Cannot validate queries on {prop!r} as field type was not reported in `/info/{endp}`"
            )

        # this is the case of a provider field
        if prop not in CONF.entry_schemas[endp]:
            if self.provider_prefix is None:
                raise ResponseError(
                    f"Found unknown field {prop!r} in `/info/{endp}` and no provider prefix was provided in `/info`"
                )
            elif not prop.startswith(f"_{self.provider_prefix}_"):
                raise ResponseError(
                    f"Found unknown field {prop!r} that did not start with provider prefix '_{self.provider_prefix}_'"
                )
            return (
                None,
                f"Found provider field {prop!r}, will not test queries as they are strictly optional.",
            )

        query_optional = (
            CONF.entry_schemas[endp].get(prop, {}).get("queryable")
            == SupportLevel.OPTIONAL
        )

        return self._construct_single_property_filters(
            prop, prop_type, sortable, endp, chosen_entry, query_optional
        )

    @staticmethod
    def _format_test_value(test_value: Any, prop_type: DataType, operator: str) -> str:
        """Formats the test value as a string according to the type of the property.

        Parameters:
            test_value: The value to format.
            prop_type: The OPTIMADE data type of the field.
            operator: The operator that will be applied to it.

        Returns:
            The value formatted as a string to use in an OPTIMADE filter.

        """
        if prop_type == DataType.LIST:
            if operator in ("HAS ALL", "HAS ANY"):
                _vals = sorted(set(test_value))
                if isinstance(test_value[0], str):
                    _vals = [f'"{val}"' for val in _vals]
                else:
                    _vals = [f"{val}" for val in _vals]
                _test_value = ",".join(_vals)
            else:
                if isinstance(test_value[0], str):
                    _test_value = f'"{test_value[0]}"'
                else:
                    _test_value = test_value[0]

        elif prop_type in (DataType.STRING, DataType.TIMESTAMP):
            _test_value = f'"{test_value}"'

        else:
            _test_value = test_value

        return _test_value

    def _construct_single_property_filters(
        self,
        prop: str,
        prop_type: DataType,
        sortable: bool,
        endp: str,
        chosen_entry: Dict[str, Any],
        query_optional: bool,
    ) -> Tuple[Optional[bool], str]:
        """This method constructs appropriate queries using all operators
        for a certain field and applies some tests:

        - inclusive operators return compatible entries, e.g. `>=` always returns
          at least the results of `=`.
        - exclusive operators never return contradictory entries, e.g.
          `nsites=1` never returns the same entries as `nsites!=1`, modulo
          pagination.

        Parameters:
            prop: The property name.
            prop_type: The property type.
            sortable: Whether the implementation has indicated that the field is sortable.
            endp: The corresponding entry endpoint.
            chosen_entry: A JSON respresentation of the chosen entry that will be used to
                construct the filters.
            query_optional: Whether to treat query success as optional.

        Returns:
            Boolean indicating success (`True`) or failure/irrelevance
            (`None`) and the string summary of the test case.

        """
        if prop == "id":
            test_value = chosen_entry.get("id")
        else:
            test_value = chosen_entry.get("attributes", {}).get(prop, "_missing")

        if test_value in ("_missing", None):
            support = CONF.entry_schemas[endp].get(prop, {}).get("support")
            queryable = CONF.entry_schemas[endp].get(prop, {}).get("queryable")
            submsg = "had no value" if test_value == "_missing" else "had `None` value"
            msg = (
                f"Chosen entry {submsg} for {prop!r} with support level {support} and queryability {queryable}, "
                f"so cannot construct test queries. This field should potentially be removed from the `/info/{endp}` endpoint response."
            )
            # None values are allowed for OPTIONAL and SHOULD, so we can just skip
            if support in (
                SupportLevel.OPTIONAL,
                SupportLevel.SHOULD,
            ):
                self._log.info(msg)
                return None, msg

            # Otherwise, None values are not allowed for MUST's, and entire missing fields are not allowed
            raise ResponseError(msg)

        using_fallback = False
        if prop_type == DataType.LIST:
            if not test_value:
                test_value = CONF.enum_fallback_values.get(endp, {}).get(prop)
                using_fallback = True

            if not test_value:
                msg = f"Not testing filters on field {prop} of type {prop_type} as no test value was found to use in filter."
                self._log.warning(msg)
                return None, msg
            if isinstance(test_value[0], dict) or isinstance(test_value[0], list):
                msg = f"Not testing filters on field {prop} of type {prop_type} with nested dictionary/list test value."
                self._log.warning(msg)
                return None, msg

            # Try to infer if the test value is a float from its string representation
            # and decide whether to do inclusive/exclusive query tests
            try:
                float(test_value[0])
                msg = f"Not testing filters on field {prop} of type {prop_type} containing float values."
                self._log.warning(msg)
                return None, msg
            except ValueError:
                pass

        if prop_type in (DataType.DICTIONARY,):
            msg = f"Not testing queries on field {prop} of type {prop_type}."
            self._log.warning(msg)
            return None, msg

        num_data_returned = {}

        inclusive_operators = CONF.inclusive_operators[prop_type]
        exclusive_operators = CONF.exclusive_operators[prop_type]

        for operator in inclusive_operators | exclusive_operators:
            # Need to pre-format list and string test values for the query
            _test_value = self._format_test_value(test_value, prop_type, operator)

            query = f"{prop} {operator} {_test_value}"
            request = f"{endp}?filter={query}"
            response, message = self._get_endpoint(
                request,
                multistage=True,
                optional=query_optional,
                expected_status_code=(200, 501),
            )

            if not response:
                if query_optional:
                    return (
                        None,
                        "Optional query {query!r} returned {reversed_response.status_code}.",
                    )
                raise ResponseError(
                    f"Unable to perform mandatory query {query!r}: responded with status code {response.status_code}"
                )

            response = response.json()

            if "meta" not in response or "more_data_available" not in response["meta"]:
                raise ResponseError(
                    f"Required field `meta->more_data_available` missing from response for {request}."
                )

            if not response["meta"]["more_data_available"]:
                num_data_returned[operator] = len(response["data"])
            else:
                num_data_returned[operator] = response["meta"].get("data_returned")

            num_response = num_data_returned[operator]

            # Numeric and string comparisons must work both ways...
            if prop_type in (
                DataType.STRING,
                DataType.INTEGER,
                DataType.FLOAT,
                DataType.TIMESTAMP,
            ) and operator not in (
                "CONTAINS",
                "STARTS",
                "STARTS WITH",
                "ENDS",
                "ENDS WITH",
            ):
                reversed_operator = operator.replace("<", ">")
                if "<" in operator:
                    reversed_operator = operator.replace("<", ">")
                elif ">" in operator:
                    reversed_operator = operator.replace(">", "<")

                # Don't try to reverse string comparison as it is ill-defined
                if prop_type == DataType.STRING and any(
                    comp in operator for comp in ("<", ">")
                ):
                    continue

                reversed_query = f"{_test_value} {reversed_operator} {prop}"
                reversed_request = f"{endp}?filter={reversed_query}"
                reversed_response, message = self._get_endpoint(
                    reversed_request,
                    multistage=True,
                    optional=query_optional,
                    expected_status_code=(200, 501),
                )

                if not reversed_response:
                    if query_optional:
                        return (
                            None,
                            "Optional query {query!r} returned {reversed_response.status_code}.",
                        )
                    raise ResponseError(
                        f"Unable to perform mandatory query {query!r}: responded with status code {reversed_response.status_code}"
                    )

                reversed_response = reversed_response.json()
                if (
                    "meta" not in reversed_response
                    or "more_data_available" not in reversed_response["meta"]
                ):
                    raise ResponseError(
                        f"Required field `meta->more_data_available` missing from response for {request}."
                    )

                if not reversed_response["meta"]["more_data_available"]:
                    num_reversed_response = len(reversed_response["data"])
                else:
                    num_reversed_response = reversed_response["meta"].get(
                        "data_returned"
                    )

                if num_response is not None and num_reversed_response is not None:
                    if reversed_response["meta"].get("data_returned") != response[
                        "meta"
                    ].get("data_returned"):
                        raise ResponseError(
                            f"Query {query} did not work both ways around: {reversed_query}, "
                            "returning different results each time."
                        )

                # check that the filter returned no entries that had a null or missing value for the filtered property
                if any(
                    entry["attributes"].get(prop, entry.get(prop, None)) is None
                    for entry in reversed_response.get("data", [])
                ):
                    raise ResponseError(
                        f"Filter {reversed_query!r} on field {prop!r} returned entries that had null or missing values for the field."
                    )

            excluded = operator in exclusive_operators
            # if we have all results on this page, check that the blessed ID is in the response
            if not response["meta"]["more_data_available"]:
                if excluded and (
                    chosen_entry["id"] in set(entry["id"] for entry in response["data"])
                ):
                    raise ResponseError(
                        f"Object {chosen_entry['id']} with value {test_value} was not excluded by {query}"
                    )

            # check that at least the archetypal structure was returned, unless we are using a fallback value
            if not excluded and not using_fallback:
                if (
                    num_data_returned[operator] is not None
                    and num_data_returned[operator] < 1
                ):
                    raise ResponseError(
                        f"Supposedly inclusive query '{query}' did not include original object ID {chosen_entry['id']} "
                        f"(with field '{prop} = {test_value}') potentially indicating a problem with filtering on this field."
                    )

            # check that the filter returned no entries that had a null or missing value for the filtered property
            if any(
                entry["attributes"].get(prop, entry.get(prop, None)) is None
                for entry in response.get("data", [])
            ):
                raise ResponseError(
                    f"Filter {query!r} on field {prop!r} returned entries that had null or missing values for the field."
                )

        if prop in CONF.unique_properties:
            if num_data_returned["="] is not None and num_data_returned["="] == 0:
                raise ResponseError(
                    f"Unable to filter field 'id' for equality, no data was returned for {query}."
                )
            if num_data_returned["="] is not None and num_data_returned["="] > 1:
                raise ResponseError(
                    f"Filter for an individual 'id' returned {num_data_returned['=']} results, when only 1 was expected."
                )

        return True, f"{prop} passed filter tests"

    def _test_info_or_links_endpoint(self, request_str: str) -> Union[bool, dict]:
        """Requests an info or links endpoint and attempts to deserialize
        the response.

        Parameters:
            request_str: The request to make, e.g. "links".

        Returns:
            `False` if the info response failed deserialization,
            otherwise returns the deserialized object.

        """
        response, _ = self._get_endpoint(request_str)
        if response:
            deserialized, _ = self._deserialize_response(
                response,
                self._response_classes[request_str],
                request=request_str,
            )
            if deserialized:
                return deserialized.dict()

        return False

    def _test_single_entry_endpoint(self, endp: str) -> None:
        """Requests and deserializes a single entry endpoint with the
        appropriate model.

        Parameters:
            request_str: The single entry request to make, e.g. "structures/id_1".

        """
        response_cls_name = endp + "/"
        if response_cls_name in self._response_classes:
            response_cls = self._response_classes[response_cls_name]
        else:
            self._log.warning(
                "Deserializing single entry response %s with generic response rather than defined endpoint.",
                endp,
            )
            response_cls = ValidatorEntryResponseOne

        response_fields = set()
        if endp in CONF.entry_schemas:
            response_fields = (
                set(CONF.entry_schemas[endp].keys())
                - CONF.top_level_non_attribute_fields
            )

        if endp in self._test_id_by_type:
            test_id = self._test_id_by_type[endp]
            request_str = f"{endp}/{test_id}"
            if response_fields:
                request_str += f"?response_fields={','.join(response_fields)}"
            response, _ = self._get_endpoint(request_str)
            if response:
                self._deserialize_response(response, response_cls, request=request_str)

    def _test_multi_entry_endpoint(self, endp: str) -> None:
        """Requests and deserializes a multi-entry endpoint with the
        appropriate model.

        TODO: deserialization is currently classed as an optional
        test until our models are robust to support levels.

        Parameters:
            request_str: The multi-entry request to make, e.g.,
                "structures?filter=nsites<10"

        """
        if endp in self._response_classes:
            response_cls = self._response_classes[endp]
        else:
            self._log.warning(
                "Deserializing multi entry response from %s with generic response rather than defined endpoint.",
                endp,
            )
            response_cls = ValidatorEntryResponseMany

        response_fields = set()
        if endp in CONF.entry_schemas:
            response_fields = (
                set(CONF.entry_schemas[endp].keys())
                - CONF.top_level_non_attribute_fields
            )

        request_str = f"{endp}?page_limit={self.page_limit}"

        if response_fields:
            request_str += f'&response_fields={",".join(response_fields)}'

        response, _ = self._get_endpoint(request_str)

        self._test_page_limit(response)

        deserialized, _ = self._deserialize_response(
            response, response_cls, request=request_str
        )

        self._get_single_id_from_multi_entry_endpoint(deserialized, request=request_str)
        if deserialized:
            self._test_data_available_matches_data_returned(
                deserialized, request=request_str
            )

    @test_case
    def _test_data_available_matches_data_returned(
        self, deserialized: Any
    ) -> Tuple[bool, str]:
        """In the case where no query is requested, `data_available`
        must equal `data_returned` in the meta response, which is tested
        here.

        Parameters:
            deserialized: The deserialized response to a multi-entry
                endpoint.

        Returns:
            `True` if successful, with a string summary.

        """
        if (
            deserialized.meta.data_available is None
            or deserialized.meta.data_returned is None
        ):
            return (
                None,
                "`meta->data_available` and/or `meta->data_returned` were not provided.",
            )

        if deserialized.meta.data_available != deserialized.meta.data_returned:
            raise ResponseError(
                "No query was performed, but `data_returned` != `data_available`."
            )

        return (
            True,
            "Meta response contained correct values for data_available and data_returned.",
        )

    def _test_versions_endpoint(self):
        """Requests and validate responses for the versions endpoint,
        which MUST exist for unversioned base URLs and MUST NOT exist
        for versioned base URLs.

        """

        # First, check that there is a versions endpoint in the appropriate place:
        # If passed a versioned URL, then strip that version from
        # the URL before looking for `/versions`.
        _old_base_url = self.base_url
        if re.match(VERSIONS_REGEXP, self.base_url_parsed.path) is not None:
            self.client.base_url = "/".join(self.client.base_url.split("/")[:-1])
            self.base_url = self.client.base_url

        response, _ = self._get_endpoint(
            CONF.versions_endpoint, expected_status_code=200
        )

        if response:
            self._test_versions_endpoint_content(
                response, request=CONF.versions_endpoint
            )

        # If passed a versioned URL, first reset the URL of the client to the
        # versioned one, then test that this versioned URL does NOT host a versions endpoint
        if re.match(VERSIONS_REGEXP, self.base_url_parsed.path) is not None:
            self.client.base_url = _old_base_url
            self.base_url = _old_base_url
            self._get_endpoint(CONF.versions_endpoint, expected_status_code=404)

    @test_case
    def _test_versions_endpoint_content(
        self, response: requests.Response
    ) -> Tuple[requests.Response, str]:
        """Checks that the response from the versions endpoint complies
        with the specification and that its 'Content-Type' header complies with
        [RFC 4180](https://tools.ietf.org/html/rfc4180.html).

        Parameters:
            response: The HTTP response from the versions endpoint.

        Raises:
            ResponseError: If any content checks fail.

        Returns:
            The successful HTTP response or `None`, and a summary string.

        """
        text_content = response.text.strip().split("\n")
        if text_content[0] != "version":
            raise ResponseError(
                f"First line of `/{CONF.versions_endpoint}` response must be 'version', not {text_content[0]!r}"
            )

        if len(text_content) <= 1:
            raise ResponseError(
                f"No version numbers found in `/{CONF.versions_endpoint}` response, only {text_content}"
            )

        for version in text_content[1:]:
            try:
                int(version)
            except ValueError:
                raise ResponseError(
                    f"Version numbers reported by `/{CONF.versions_endpoint}` must be integers specifying the major version, not {text_content}."
                )

        content_type = response.headers.get("content-type")
        if not content_type:
            raise ResponseError(
                "Missing 'Content-Type' in response header from `/versions`."
            )

        content_type = [_.replace(" ", "") for _ in content_type.split(";")]

        self._test_versions_headers(
            content_type,
            ("text/csv", "text/plain"),
            optional=True,
            request=CONF.versions_endpoint,
        )
        self._test_versions_headers(
            content_type,
            "header=present",
            optional=True,
            request=CONF.versions_endpoint,
        )

        return response, "`/versions` endpoint responded correctly."

    @test_case
    def _test_versions_headers(
        self,
        content_type: Dict[str, Any],
        expected_parameter: Union[str, List[str]],
    ) -> Tuple[Dict[str, Any], str]:
        """Tests that the `Content-Type` field of the `/versions` header contains
        the passed parameter.

        Arguments:
            content_type: The 'Content-Type' field from the response of the `/versions` endpoint.
            expected_paramter: A substring or list of substrings that are expected in
                the Content-Type of the response. If multiple strings are passed, they will
                be treated as possible alternatives to one another.

        Raises:
            ResponseError: If the expected 'Content-Type' parameter is missing.

        Returns:
            The HTTP response headers and a summary string.

        """

        if isinstance(expected_parameter, str):
            expected_parameter = [expected_parameter]

        if not any(param in content_type for param in expected_parameter):
            raise ResponseError(
                f"Incorrect 'Content-Type' header {';'.join(content_type)!r}.\n"
                f"Missing at least one expected parameter(s): {expected_parameter!r}"
            )

        return (
            content_type,
            f"`/versions` response had one of the expected Content-Type parameters {expected_parameter}.",
        )

    def _test_as_type(self) -> None:
        """Tests that the base URL of the validator (i.e. with no
        additional path added) validates with the model selected.

        """
        response, _ = self._get_endpoint("")
        if response:
            self._log.debug("Deserialzing response as type %s", self.as_type_cls)
            self._deserialize_response(response, self.as_type_cls)

    def _test_bad_version_returns_553(self) -> None:
        """Tests that a garbage version number responds with a 553
        error code.

        """
        expected_status_code = 553
        if re.match(VERSIONS_REGEXP, self.base_url_parsed.path) is not None:
            expected_status_code = [404, 400]

        self._get_endpoint(
            "v123123", expected_status_code=expected_status_code, optional=True
        )

    @test_case
    def _test_page_limit(
        self, response: requests.models.Response, check_next_link: int = 5
    ) -> Tuple[Optional[bool], str]:
        """Test that a multi-entry endpoint obeys the page limit by
        following pagination links up to a depth of `check_next_link`.

        Parameters:
            response: The response to test for page limit
                compliance.
            check_next_link: Maximum recursion depth for following
                pagination links.

        Returns:
            `True` if the test was successful and `None` if not, with a
            string summary.

        """
        try:
            response = response.json()
        except (AttributeError, json.JSONDecodeError):
            raise ResponseError("Unable to test endpoint `page_limit` parameter.")

        try:
            num_entries = len(response["data"])
        except (KeyError, TypeError):
            raise ResponseError(
                "Response under `data` field was missing or had wrong type."
            )

        if num_entries > self.page_limit:
            raise ResponseError(
                f"Endpoint did not obey page limit: {num_entries} entries vs {self.page_limit} limit"
            )

        try:
            more_data_available = response["meta"]["more_data_available"]
        except KeyError:
            raise ResponseError("Field `meta->more_data_available` was missing.")

        if more_data_available and check_next_link:
            try:
                next_link = response["links"]["next"]
                if isinstance(next_link, dict):
                    next_link = next_link["href"]
            except KeyError:
                raise ResponseError(
                    "Endpoint suggested more data was available but provided no valid links->next link."
                )

            if not isinstance(next_link, str):
                raise ResponseError(
                    f"Unable to parse links->next {next_link!r} as a link."
                )

            self._log.debug("Following pagination link to %r.", next_link)
            next_response, _ = self._get_endpoint(next_link)
            if not next_response:
                raise ResponseError(
                    f"Error when testing pagination: the response from `links->next` {next_link!r} failed the previous test."
                )

            check_next_link = bool(check_next_link - 1)
            self._test_page_limit(
                next_response,
                check_next_link=check_next_link,
                multistage=check_next_link,
            )

        return (
            True,
            f"Endpoint obeyed page limit of {self.page_limit} by returning {num_entries} entries.",
        )

    @test_case
    def _get_single_id_from_multi_entry_endpoint(self, deserialized):
        """Scrape an ID from the multi-entry endpoint to use as query
        for single entry endpoint.

        """
        if deserialized and deserialized.data:
            self._test_id_by_type[deserialized.data[0].type] = deserialized.data[0].id
            self._log.debug(
                "Set type %s test ID to %s",
                deserialized.data[0].type,
                deserialized.data[0].id,
            )
        else:
            return (
                None,
                "No entries found under endpoint to scrape ID from. "
                "This may be caused by previous errors, if e.g. the endpoint failed deserialization.",
            )
        return (
            self._test_id_by_type[deserialized.data[0].type],
            f"successfully scraped test ID from {deserialized.data[0].type} endpoint",
        )

    @test_case
    def _deserialize_response(
        self, response: requests.models.Response, response_cls: Any, request: str = None
    ) -> Tuple[Any, str]:
        """Try to create the appropriate pydantic model from the response.

        Parameters:
            response: The response to try to deserialize.
            response_cls: The class to use for deserialization.
            request: Optional string that will be displayed as the attempted
                request in the validator output.

        Returns:
            The deserialized object (or `None` if unsuccessful) and a
                human-readable summary

        """
        if not response:
            raise ResponseError("Request failed")
        try:
            json_response = response.json()
        except json.JSONDecodeError:
            raise ResponseError(
                f"Unable to decode response as JSON. Response: {response}"
            )

        self._log.debug(
            f"Deserializing {json.dumps(json_response, indent=2)} as model {response_cls}"
        )

        return (
            response_cls(**json_response),
            "deserialized correctly as object of type {}".format(response_cls),
        )

    @test_case
    def _get_available_endpoints(
        self, base_info: Union[Any, Dict[str, Any]]
    ) -> Tuple[Optional[List[str]], str]:
        """Tries to get `entry_types_by_format` from base info response
        even if it could not be deserialized.

        Parameters:
            base_info: Either the unvalidated JSON representation of the
                base info, or the deserialized object.

        Returns:
            The list of JSON entry endpoints (or `None` if unavailable)
            and a string summary.

        """
        for _ in [0]:
            try:
                available_json_entry_endpoints = base_info["data"]["attributes"][
                    "entry_types_by_format"
                ]["json"]
                break
            except (KeyError, TypeError):
                raise ResponseError(
                    "Unable to get entry_types_by_format from unserializable base info response {}.".format(
                        base_info
                    )
                )
        else:
            raise ResponseError(
                "Unable to find any JSON entry types in entry_types_by_format"
            )

        if self.index and available_json_entry_endpoints != []:
            raise ResponseError(
                "No entry endpoint are allowed for an Index meta-database"
            )

        for non_entry_endpoint in CONF.non_entry_endpoints:
            if non_entry_endpoint in available_json_entry_endpoints:
                raise ResponseError(
                    f'Illegal entry "{non_entry_endpoint}" was found in entry_types_by_format"'
                )

        # Filter out custom extension endpoints that are not covered in the specification
        available_json_entry_endpoints = [
            endp
            for endp in available_json_entry_endpoints
            if endp in CONF.entry_endpoints
        ]

        return (
            available_json_entry_endpoints,
            "successfully found available entry types in baseinfo",
        )

    @test_case
    def _get_endpoint(
        self, request_str: str, expected_status_code: Union[List[int], int] = 200
    ) -> Tuple[Optional[requests.Response], str]:
        """Gets the response from the endpoint specified by `request_str`.
        function is wrapped by the `test_case` decorator

        Parameters:
            request_str: The request to make to the client.
            expected_status_code: If the request responds with a different
                status code to this one, raise a ResponseError.

        Returns:
            The response to the request (if successful) or `None`, plus
            a string summary.

        """
        request_str = request_str.replace("\n", "")
        response = self.client.get(request_str)

        if isinstance(expected_status_code, int):
            expected_status_code = [expected_status_code]

        if response.status_code not in expected_status_code:
            message = f"Request to '{request_str}' returned HTTP code {response.status_code} and not the allowed {expected_status_code}."
            message += "\nAdditional details:"
            try:
                for error in response.json().get("errors", []):
                    message += f'\n  {error.get("title", "N/A")}: {error.get("detail", "N/A")} ({error.get("source", {}).get("pointer", "N/A")})'
            except json.JSONDecodeError:
                message += f"\n  Could not parse response as JSON. Content type was {response.headers.get('content-type')!r}."
            raise ResponseError(message)

        return response, f"received expected response: {response}."



from optimade.server.mappers.entries import BaseResourceMapper
from optimade.models.structures import StructureResource

__all__ = ("StructureMapper",)


class StructureMapper(BaseResourceMapper):

    LENGTH_ALIASES = (
        ("elements", "nelements"),
        ("element_ratios", "nelements"),
        ("cartesian_site_positions", "nsites"),
        ("species_at_sites", "nsites"),
    )
    ENTRY_RESOURCE_CLASS = StructureResource



# pylint: disable=no-self-argument
from enum import Enum

from pydantic import (  # pylint: disable=no-name-in-module
    AnyUrl,
    root_validator,
)
from typing import Union, Optional

from optimade.models.jsonapi import Link, Attributes
from optimade.models.entries import EntryResource
from optimade.models.utils import StrictField


__all__ = (
    "LinksResourceAttributes",
    "LinksResource",
)


class LinkType(Enum):
    """Enumeration of link_type values"""

    CHILD = "child"
    ROOT = "root"
    EXTERNAL = "external"
    PROVIDERS = "providers"


class Aggregate(Enum):
    """Enumeration of aggregate values"""

    OK = "ok"
    TEST = "test"
    STAGING = "staging"
    NO = "no"


class LinksResourceAttributes(Attributes):
    """Links endpoint resource object attributes"""

    name: str = StrictField(
        ...,
        description="Human-readable name for the OPTIMADE API implementation, e.g., for use in clients to show the name to the end-user.",
    )
    description: str = StrictField(
        ...,
        description="Human-readable description for the OPTIMADE API implementation, e.g., for use in clients to show a description to the end-user.",
    )
    base_url: Optional[Union[AnyUrl, Link]] = StrictField(
        ...,
        description="JSON API links object, pointing to the base URL for this implementation",
    )

    homepage: Optional[Union[AnyUrl, Link]] = StrictField(
        ...,
        description="JSON API links object, pointing to a homepage URL for this implementation",
    )

    link_type: LinkType = StrictField(
        ...,
        title="Link Type",
        description="""The type of the linked relation.
MUST be one of these values: 'child', 'root', 'external', 'providers'.""",
    )

    aggregate: Optional[Aggregate] = StrictField(
        Aggregate.OK,
        title="Aggregate",
        description="""A string indicating whether a client that is following links to aggregate results from different OPTIMADE implementations should follow this link or not.
This flag SHOULD NOT be indicated for links where `link_type` is not `child`.

If not specified, clients MAY assume that the value is `ok`.
If specified, and the value is anything different than `ok`, the client MUST assume that the server is suggesting not to follow the link during aggregation by default (also if the value is not among the known ones, in case a future specification adds new accepted values).

Specific values indicate the reason why the server is providing the suggestion.
A client MAY follow the link anyway if it has reason to do so (e.g., if the client is looking for all test databases, it MAY follow the links marked with `aggregate`=`test`).

If specified, it MUST be one of the values listed in section Link Aggregate Options.""",
    )

    no_aggregate_reason: Optional[str] = StrictField(
        None,
        description="""An OPTIONAL human-readable string indicating the reason for suggesting not to aggregate results following the link.
It SHOULD NOT be present if `aggregate`=`ok`.""",
    )


class LinksResource(EntryResource):
    """A Links endpoint resource object"""

    type: str = StrictField(
        "links",
        const="links",
        description="These objects are described in detail in the section Links Endpoint",
        pattern="^links$",
    )

    attributes: LinksResourceAttributes = StrictField(
        ...,
        description="A dictionary containing key-value pairs representing the Links resource's properties.",
    )

    @root_validator(pre=True)
    def relationships_must_not_be_present(cls, values):
        if values.get("relationships", None) is not None:
            raise ValueError('"relationships" is not allowed for links resources')
        return values



from fastapi import APIRouter, Depends, Request

from optimade.models import LinksResponse, LinksResource
from optimade.server.config import CONFIG
from optimade.server.entry_collections import create_collection
from optimade.server.mappers import LinksMapper
from optimade.server.query_params import EntryListingQueryParams
from optimade.server.routers.utils import get_entries
from optimade.server.schemas import ERROR_RESPONSES

router = APIRouter(redirect_slashes=True)

links_coll = create_collection(
    name=CONFIG.links_collection,
    resource_cls=LinksResource,
    resource_mapper=LinksMapper,
)


@router.get(
    "/links",
    response_model=LinksResponse,
    response_model_exclude_unset=True,
    tags=["Links"],
    responses=ERROR_RESPONSES,
)
def get_links(
    request: Request, params: EntryListingQueryParams = Depends()
) -> LinksResponse:
    return get_entries(
        collection=links_coll, response=LinksResponse, request=request, params=params
    )



"""This submodule implements the
[`MongoTransformer`][optimade.filtertransformers.mongo.MongoTransformer],
which takes the parsed filter and converts it to a valid pymongo/BSON query.
"""


import copy
import warnings
from typing import Dict, Any
from lark import v_args, Token
from optimade.filtertransformers.base_transformer import BaseTransformer, Quantity


from optimade.server.warnings import TimestampNotRFCCompliant

__all__ = ("MongoTransformer",)


class MongoTransformer(BaseTransformer):
    """A filter transformer for the MongoDB backend.

    Parses a lark tree into a dictionary representation to be
    used by pymongo or mongomock. Uses post-processing functions
    to handle some specific edge-cases for MongoDB.

    """

    operator_map = {
        "<": "$lt",
        "<=": "$lte",
        ">": "$gt",
        ">=": "$gte",
        "!=": "$ne",
        "=": "$eq",
    }

    def postprocess(self, query: Dict[str, Any]):
        """Used to post-process the nested dictionary of the parsed query."""
        query = self._apply_relationship_filtering(query)
        query = self._apply_length_operators(query)
        query = self._apply_unknown_or_null_filter(query)
        query = self._apply_mongo_id_filter(query)
        query = self._apply_mongo_date_filter(query)
        return query

    def value_list(self, arg):
        # value_list: [ OPERATOR ] value ( "," [ OPERATOR ] value )*
        # NOTE: no support for optional OPERATOR, yet, so this takes the
        # parsed values and returns an error if that is being attempted
        for value in arg:
            if str(value) in self.operator_map.keys():
                raise NotImplementedError(
                    f"OPERATOR {value} inside value_list {arg} not implemented."
                )

        return arg

    def value_zip(self, arg):
        # value_zip: [ OPERATOR ] value ":" [ OPERATOR ] value (":" [ OPERATOR ] value)*
        raise NotImplementedError("Correlated list queries are not supported.")

    def value_zip_list(self, arg):
        # value_zip_list: value_zip ( "," value_zip )*
        raise NotImplementedError("Correlated list queries are not supported.")

    def expression(self, arg):
        # expression: expression_clause ( OR expression_clause )
        # expression with and without 'OR'
        return {"$or": arg} if len(arg) > 1 else arg[0]

    def expression_clause(self, arg):
        # expression_clause: expression_phrase ( AND expression_phrase )*
        # expression_clause with and without 'AND'
        return {"$and": arg} if len(arg) > 1 else arg[0]

    def expression_phrase(self, arg):
        # expression_phrase: [ NOT ] ( comparison | "(" expression ")" )
        return self._recursive_expression_phrase(arg)

    @v_args(inline=True)
    def property_first_comparison(self, quantity, query):
        # property_first_comparison: property ( value_op_rhs | known_op_rhs | fuzzy_string_op_rhs | set_op_rhs |
        # set_zip_op_rhs | length_op_rhs )

        # Awkwardly, MongoDB will match null fields in $ne filters,
        # so we need to add a check for null equality in evey $ne query.
        if "$ne" in query:
            return {"$and": [{quantity: query}, {quantity: {"$ne": None}}]}

        # Check if a $size query is being made (indicating a length_op_rhs filter); if so, check for
        # a defined length alias to replace the $size call with the corresponding filter on the
        # length quantity then carefully merge the two queries.
        #
        # e.g. `("elements", {"$size": 2, "$all": ["Ag", "Au"]})` should become
        # `{"elements": {"$all": ["Ag", "Au"]}, "nelements": 2}` if the `elements` -> `nelements`
        # length alias is defined.
        if "$size" in query:
            if (
                getattr(self.backend_mapping.get(quantity), "length_quantity", None)
                is not None
            ):
                size_query = {
                    self.backend_mapping[
                        quantity
                    ].length_quantity.backend_field: query.pop("$size")
                }

                final_query = {}
                if query:
                    final_query = {quantity: query}
                for q in size_query:
                    if q in final_query:
                        final_query[q].update(size_query[q])
                    else:
                        final_query[q] = size_query[q]

                return final_query

        return {quantity: query}

    def constant_first_comparison(self, arg):
        # constant_first_comparison: constant OPERATOR ( non_string_value | not_implemented_string )
        return {
            arg[2]: {self.operator_map[self._reversed_operator_map[arg[1]]]: arg[0]}
        }

    @v_args(inline=True)
    def value_op_rhs(self, operator, value):
        # value_op_rhs: OPERATOR value
        return {self.operator_map[operator]: value}

    def known_op_rhs(self, arg):
        # known_op_rhs: IS ( KNOWN | UNKNOWN )
        # The OPTIMADE spec also required a type comparison with null, this must be post-processed
        # so here we use a special key "#known" which will get replaced in post-processing with the
        # expanded dict
        return {"#known": arg[1] == "KNOWN"}

    def fuzzy_string_op_rhs(self, arg):
        # fuzzy_string_op_rhs: CONTAINS value | STARTS [ WITH ] value | ENDS [ WITH ] value

        # The WITH keyword may be omitted.
        if isinstance(arg[1], Token) and arg[1].type == "WITH":
            pattern = arg[2]
        else:
            pattern = arg[1]

        # CONTAINS
        if arg[0] == "CONTAINS":
            regex = f"{pattern}"
        elif arg[0] == "STARTS":
            regex = f"^{pattern}"
        elif arg[0] == "ENDS":
            regex = f"{pattern}$"
        return {"$regex": regex}

    def set_op_rhs(self, arg):
        # set_op_rhs: HAS ( [ OPERATOR ] value | ALL value_list | ANY value_list | ONLY value_list )

        if len(arg) == 2:
            # only value without OPERATOR
            return {"$in": arg[1:]}

        if arg[1] == "ALL":
            return {"$all": arg[2]}

        if arg[1] == "ANY":
            return {"$in": arg[2]}

        if arg[1] == "ONLY":
            return {"$all": arg[2], "$size": len(arg[2])}

        # value with OPERATOR
        raise NotImplementedError(
            f"set_op_rhs not implemented for use with OPERATOR. Given: {arg}"
        )

    def property(self, args):
        # property: IDENTIFIER ( "." IDENTIFIER )*
        quantity = super().property(args)
        if isinstance(quantity, Quantity):
            quantity = quantity.backend_field

        return ".".join([quantity] + args[1:])

    def length_op_rhs(self, arg):
        # length_op_rhs: LENGTH [ OPERATOR ] value
        if len(arg) == 2 or (len(arg) == 3 and arg[1] == "="):
            return {"$size": arg[-1]}

        if arg[1] in self.operator_map and arg[1] != "!=":
            # create an invalid query that needs to be post-processed
            # e.g. {'$size': {'$gt': 2}}, which is not allowed by Mongo.
            return {"$size": {self.operator_map[arg[1]]: arg[-1]}}

        raise NotImplementedError(
            f"Operator {arg[1]} not implemented for LENGTH filter."
        )

    def set_zip_op_rhs(self, arg):
        # set_zip_op_rhs: property_zip_addon HAS ( value_zip | ONLY value_zip_list | ALL value_zip_list |
        # ANY value_zip_list )
        raise NotImplementedError("Correlated list queries are not supported.")

    def property_zip_addon(self, arg):
        # property_zip_addon: ":" property (":" property)*
        raise NotImplementedError("Correlated list queries are not supported.")

    def _recursive_expression_phrase(self, arg):
        """Helper function for parsing `expression_phrase`. Recursively sorts out
        the correct precedence for `$and`, `$or` and `$nor`.

        """
        if len(arg) == 1:
            # without NOT
            return arg[0]

        # handle the case of {"$not": {"$or": [expr1, expr2]}} using {"$nor": [expr1, expr2]}.
        if "$or" in arg[1]:
            return {"$nor": self._recursive_expression_phrase([arg[1]["$or"]])}

        # handle the case of {"$not": {"$and": [expr1, expr2]}} using per-expression negation,
        # e.g. {"$and": [{prop1: {"$not": expr1}}, {prop2: {"$not": ~expr2}}]}.
        # Note that this is not the same as NOT (expr1 AND expr2)!
        if "$and" in arg[1]:
            return {
                "$and": [
                    self._recursive_expression_phrase(["NOT", subdict])
                    for subdict in arg[1]["$and"]
                ]
            }

        # simple case of negating one expression, from NOT (expr) to ~expr.
        return {prop: {"$not": expr} for prop, expr in arg[1].items()}

    def _apply_length_operators(self, filter_: dict) -> dict:
        """Check for any invalid pymongo queries that involve applying a
        comparison operator to the length of a field, and transform
        them into a test for existence of the relevant entry, e.g.
        "list LENGTH > 3" becomes "does the 4th list entry exist?".

        """

        def check_for_length_op_filter(_, expr):
            return (
                isinstance(expr, dict)
                and "$size" in expr
                and isinstance(expr["$size"], dict)
            )

        def apply_length_op(subdict, prop, expr):
            # assumes that the dictionary only has one element by design
            # (we just made it above in the transformer)
            operator, value = list(expr["$size"].items())[0]
            if operator in self.operator_map.values() and operator != "$ne":
                # worth being explicit here, I think
                _prop = None
                existence = None
                if operator == "$gt":
                    _prop = f"{prop}.{value + 1}"
                    existence = True
                elif operator == "$gte":
                    _prop = f"{prop}.{value}"
                    existence = True
                elif operator == "$lt":
                    _prop = f"{prop}.{value}"
                    existence = False
                elif operator == "$lte":
                    _prop = f"{prop}.{value + 1}"
                    existence = False
                if _prop is not None:
                    subdict.pop(prop)
                    subdict[_prop] = {"$exists": existence}

            return subdict

        return recursive_postprocessing(
            filter_,
            check_for_length_op_filter,
            apply_length_op,
        )

    def _apply_relationship_filtering(self, filter_: dict) -> dict:
        """Check query for property names that match the entry
        types, and transform them as relationship filters rather than
        property filters.

        """

        def check_for_entry_type(prop, _):
            return str(prop).count(".") == 1 and str(prop).split(".")[0] in (
                "structures",
                "references",
            )

        def replace_with_relationship(subdict, prop, expr):
            _prop, _field = str(prop).split(".")
            if _field != "id":
                raise NotImplementedError(
                    f'Cannot filter relationships by field "{_field}", only "id" is supported.'
                )

            # in the case of HAS ONLY, the size operator needs to be applied
            # one level up, i.e. excluding the field
            if "$size" in expr:
                if "$and" not in subdict:
                    subdict["$and"] = []
                subdict["$and"].extend(
                    [
                        {
                            f"relationships.{_prop}.data": {
                                "$size": expr.pop("$size"),
                            }
                        },
                        {f"relationships.{_prop}.data.{_field}": expr},
                    ]
                )
            else:
                subdict[f"relationships.{_prop}.data.{_field}"] = expr

            subdict.pop(prop)

            return subdict

        return recursive_postprocessing(
            filter_, check_for_entry_type, replace_with_relationship
        )

    def _apply_unknown_or_null_filter(self, filter_: dict) -> dict:
        """This method loops through the query and replaces the check for
        KNOWN with a check for existence and a check for not null, and the
        inverse for UNKNOWN.

        """

        def check_for_known_filter(_, expr):
            """Find cases where the query dict looks like
            `{"field": {"#known": T/F}}` or
            `{"field": "$not": {"#known": T/F}}`, which is a magic word
            for KNOWN/UNKNOWN filters in this transformer.

            """
            return isinstance(expr, dict) and (
                "#known" in expr or "#known" in expr.get("$not", {})
            )

        def replace_known_filter_with_or(subdict, prop, expr):
            """Replace magic key `"#known"` (added by this transformer) with the appropriate
            combination of `$exists` and/or test for nullity.
            combination of $exists and/or $eq/$ne null.

            """
            not_ = set(expr.keys()) == {"$not"}
            if not_:
                expr = expr["$not"]

            exists = expr["#known"] ^ not_

            top_level_key = "$or"
            comparison_operator = "$eq"
            if exists:
                top_level_key = "$and"
                comparison_operator = "$ne"

            if top_level_key not in subdict:
                subdict[top_level_key] = []

            subdict[top_level_key].append({prop: {"$exists": exists}})
            subdict[top_level_key].append({prop: {comparison_operator: None}})

            subdict.pop(prop)

            return subdict

        return recursive_postprocessing(
            filter_, check_for_known_filter, replace_known_filter_with_or
        )

    def _apply_mongo_id_filter(self, filter_: dict) -> dict:
        """This method loops through the query and replaces any operations
        on the special Mongodb `_id` key with the corresponding operation
        on a BSON `ObjectId` type.
        """

        def check_for_id_key(prop, _):
            """Find cases where the query dict is operating on the `_id` field."""
            return prop == "_id"

        def replace_str_id_with_objectid(subdict, prop, expr):
            from bson import ObjectId

            for operator in subdict[prop]:
                val = subdict[prop][operator]
                if operator not in ("$eq", "$ne"):
                    if self.mapper is not None:
                        prop = self.mapper.get_optimade_field(prop)
                    raise NotImplementedError(
                        f"Operator {operator} not supported for query on field {prop!r}, can only test for equality"
                    )
                if isinstance(val, str):
                    subdict[prop][operator] = ObjectId(val)
            return subdict

        return recursive_postprocessing(
            filter_, check_for_id_key, replace_str_id_with_objectid
        )

    def _apply_mongo_date_filter(self, filter_: dict) -> dict:
        """This method loops through the query and replaces any operations
        on suspected timestamp properties with the corresponding operation
        on a BSON `DateTime` type.
        """

        def check_for_timestamp_field(prop, _):
            """Find cases where the query dict is operating on a timestamp field."""
            if self.mapper is not None:
                prop = self.mapper.get_optimade_field(prop)
            return prop == "last_modified"

        def replace_str_date_with_datetime(subdict, prop, expr):
            """Encode suspected dates in with BSON."""
            import bson.json_util

            for operator in subdict[prop]:
                query_datetime = bson.json_util.loads(
                    bson.json_util.dumps({"$date": subdict[prop][operator]}),
                    json_options=bson.json_util.DEFAULT_JSON_OPTIONS.with_options(
                        tz_aware=True, tzinfo=bson.tz_util.utc
                    ),
                )
                if query_datetime.microsecond != 0:
                    warnings.warn(
                        f"Query for timestamp {subdict[prop][operator]!r} for field {prop!r} contained microseconds, which is not RFC3339 compliant. "
                        "This may cause undefined behaviour for the underlying database.",
                        TimestampNotRFCCompliant,
                    )

                subdict[prop][operator] = query_datetime

            return subdict

        return recursive_postprocessing(
            filter_, check_for_timestamp_field, replace_str_date_with_datetime
        )


def recursive_postprocessing(filter_, condition, replacement):
    """Recursively descend into the query, checking each dictionary
    (contained in a list, or as an entry in another dictionary) for
    the condition passed. If the condition is true, apply the
    replacement to the dictionary.

    Parameters:
        filter_ (list/dict): the filter_ to process.
        condition (callable): a function that returns True if the
            replacement function should be applied. It should take
            as arguments the property and expression from the filter_,
            as would be returned by iterating over `filter_.items()`.
        replacement (callable): a function that returns the processed
            dictionary. It should take as arguments the dictionary
            to modify, the property and the expression (as described
            above).

    Example:
        For the simple case of replacing one field name with
        another, the following functions could be used:

        ```python
        def condition(prop, expr):
            return prop == "field_name_old"

        def replacement(d, prop, expr):
            d["field_name_old"] = d.pop(prop)

        filter_ = recursive_postprocessing(
            filter_, condition, replacement
        )

        ```

    """
    if isinstance(filter_, list):
        result = [recursive_postprocessing(q, condition, replacement) for q in filter_]
        return result

    if isinstance(filter_, dict):
        # this could potentially lead to memory leaks if the filter_ is *heavily* nested
        _cached_filter = copy.deepcopy(filter_)
        for prop, expr in filter_.items():
            if condition(prop, expr):
                _cached_filter = replacement(_cached_filter, prop, expr)
            elif isinstance(expr, list):
                _cached_filter[prop] = [
                    recursive_postprocessing(q, condition, replacement) for q in expr
                ]
        return _cached_filter

    return filter_



"""
The base for all adapters.

An entry resource adapter is a tool to wrap OPTIMADE JSON-deserialized Python dictionaries
in the relevant pydantic model for the particular resource.

This means data resources in an OPTIMADE REST API response can be converted to valid Python
types written specifically for them.
One can then use the standard pydantic functionality on the wrapped objects, reasoning about
the embedded hierarchical types as well as retrieve default values for properties not supplied
by the raw API response resource.

Furthermore, the entry resource adapter allows conversion between the entry resource and any
implemented equivalent data structure.

See [`Reference`][optimade.adapters.references.adapter.Reference] and
[`Structure`][optimade.adapters.structures.adapter.Structure] to find out what the entry
resources can be converted to for [`ReferenceResource`][optimade.models.references.ReferenceResource]s
and [`StructureResource`][optimade.models.structures.StructureResource]s, respectively.
"""
import re
from typing import Union, Dict, Callable, Any, Tuple, List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from optimade.models import EntryResource

from optimade.adapters.logger import LOGGER


class EntryAdapter:
    """
    Base class for lazy resource entry adapters.

    Attributes:
        ENTRY_RESOURCE (EntryResource): Entry resource to store entry as.
        _type_converters (Dict[str, Callable]): Dictionary of valid conversion types for entry.
        as_<_type_converters>: Convert entry to a type listed in `_type_converters`.

    """

    ENTRY_RESOURCE: EntryResource = EntryResource
    _type_converters: Dict[str, Callable] = {}

    def __init__(self, entry: dict) -> None:
        """
        Parameters:
            entry (dict): A JSON OPTIMADE single resource entry.
        """
        self._entry = None
        self._converted = {}

        self.entry = entry

        # Note that these return also the default values for otherwise non-provided properties.
        self._common_converters = {
            "json": self.entry.json,  # Return JSON serialized string, see https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeljson
            "dict": self.entry.dict,  # Return Python dict, see https://pydantic-docs.helpmanual.io/usage/exporting_models/#modeldict
        }

    @property
    def entry(self) -> EntryResource:
        """Get OPTIMADE entry.

        Returns:
            The entry resource.

        """
        return self._entry

    @entry.setter
    def entry(self, value: dict) -> None:
        """Set OPTIMADE entry.

        If already set, print that this can _only_ be set once.

        Parameters:
            value (dict): Raw entry to wrap in the relevant pydantic model represented by `ENTRY_RESOURCE`.

        """
        if self._entry is None:
            self._entry = self.ENTRY_RESOURCE(**value)
        else:
            LOGGER.warning("entry can only be set once and is already set.")

    def convert(self, format: str) -> Any:
        """Convert OPTIMADE entry to desired format.

        Parameters:
            format (str): Type or format to which the entry should be converted.

        Raises:
            AttributeError: If `format` can not be found in `_type_converters` or `_common_converters`.

        Returns:
            The converted entry according to the desired format or type.

        """
        if (
            format not in self._type_converters
            and format not in self._common_converters
        ):
            raise AttributeError(
                f"Non-valid entry type to convert to: {format}\n"
                f"Valid entry types: {tuple(self._type_converters.keys()) + tuple(self._common_converters.keys())}"
            )

        if self._converted.get(format, None) is None:
            if format in self._type_converters:
                self._converted[format] = self._type_converters[format](self.entry)
            else:
                self._converted[format] = self._common_converters[format]()

        return self._converted[format]

    @staticmethod
    def _get_model_attributes(
        starting_instances: Union[Tuple[BaseModel], List[BaseModel]], name: str
    ) -> Any:
        """Helper method for retrieving the OPTIMADE model's attribute, supporting "."-nested attributes"""
        for res in starting_instances:
            nested_attributes = name.split(".")
            for nested_attribute in nested_attributes:
                if nested_attribute in getattr(res, "__fields__", {}):
                    res = getattr(res, nested_attribute)
                else:
                    res = None
                    break
            if res is not None:
                return res
        raise AttributeError

    def __getattr__(self, name: str) -> Any:
        """Get converted entry or attribute from OPTIMADE entry.

        Support any level of "."-nested OPTIMADE `ENTRY_RESOURCE` attributes, e.g.,
        `attributes.species` for [`StuctureResource`][optimade.models.structures.StructureResource].

        Note:
            All nested attributes must individually be subclasses of `pydantic.BaseModel`,
            i.e., one can not access nested attributes in lists by passing a "."-nested `name` to this method,
            e.g., `attributes.species.name` or `attributes.species[0].name` will not work for variable `name`.

        Order:

        - Try to return converted entry if using `as_<_type_converters key>`.
        - Try to return OPTIMADE `ENTRY_RESOURCE` (nested) attribute.
        - Try to return OPTIMADE `ENTRY_RESOURCE.attributes` (nested) attribute.
        - Raise `AttributeError`.

        Parameters:
            name (str): Requested attribute.

        Raises:
            AttributeError: If the requested attribute is not recognized.
                See above for the description of the order in which an attribute is tested for validity.

        """
        # as_<entry_type>
        if name.startswith("as_"):
            entry_type = "_".join(name.split("_")[1:])
            return self.convert(entry_type)

        # Try returning ENTRY_RESOURCE attribute
        try:
            res = self._get_model_attributes((self.entry, self.entry.attributes), name)
        except AttributeError:
            pass
        else:
            return res

        # Non-valid attribute
        entry_resource_name = re.match(
            r"(<class ')([a-zA-Z_]+\.)*([a-zA-Z_]+)('>)", str(self.ENTRY_RESOURCE)
        )
        entry_resource_name = (
            entry_resource_name.group(3)
            if entry_resource_name is not None
            else "UNKNOWN RESOURCE"
        )
        raise AttributeError(
            f"Unknown attribute: {name}\n"
            "If you want to get a converted entry as <entry_type> use `as_<entry_type>`, "
            f"where `<entry_type>` is one of {tuple(self._type_converters.keys()) + tuple(self._common_converters.keys())}\n"
            f"Otherwise, you can try to retrieve an OPTIMADE {entry_resource_name} attribute or property."
        )



"""Logging to both file and terminal"""
import logging
import os
from pathlib import Path
import sys

from uvicorn.logging import DefaultFormatter


# Instantiate LOGGER
LOGGER = logging.getLogger("optimade")
LOGGER.setLevel(logging.DEBUG)

# Handler
CONSOLE_HANDLER = logging.StreamHandler(sys.stdout)
try:
    from optimade.server.config import CONFIG

    CONSOLE_HANDLER.setLevel(CONFIG.log_level.value.upper())

    if CONFIG.debug:
        CONSOLE_HANDLER.setLevel(logging.DEBUG)


except ImportError:
    CONSOLE_HANDLER.setLevel(os.getenv("OPTIMADE_LOG_LEVEL", "INFO").upper())

# Formatter
CONSOLE_FORMATTER = DefaultFormatter("%(levelprefix)s [%(name)s] %(message)s")
CONSOLE_HANDLER.setFormatter(CONSOLE_FORMATTER)

# Add handler to LOGGER
LOGGER.addHandler(CONSOLE_HANDLER)

# Save a file with all messages (DEBUG level)
try:
    from optimade.server.config import CONFIG

    LOGS_DIR = CONFIG.log_dir
except ImportError:
    LOGS_DIR = Path(os.getenv("OPTIMADE_LOG_DIR", "/var/log/optimade/")).resolve()

try:
    LOGS_DIR.mkdir(exist_ok=True)

    # Handlers
    FILE_HANDLER = logging.handlers.RotatingFileHandler(
        LOGS_DIR.joinpath("optimade.log"), maxBytes=1000000, backupCount=5
    )

except OSError:
    LOGGER.warning(
        f"""Log files are not saved.

    This is usually due to not being able to access a specified log folder or write to files
    in the specified log location, i.e., a `PermissionError` has been raised.

    To solve this, either set the OPTIMADE_LOG_DIR environment variable to a location
    you have permission to write to or create the {LOGS_DIR} folder, which is
    the default logging folder, with write permissions for the Unix user running the server.
    """
    )
else:
    FILE_HANDLER.setLevel(logging.DEBUG)

    # Formatter
    FILE_FORMATTER = logging.Formatter(
        "[%(levelname)-8s %(asctime)s %(filename)s:%(lineno)d][%(name)s] %(message)s",
        "%d-%m-%Y %H:%M:%S",
    )
    FILE_HANDLER.setFormatter(FILE_FORMATTER)

    # Add handler to LOGGER
    LOGGER.addHandler(FILE_HANDLER)



from fastapi import HTTPException

__all__ = (
    "BadRequest",
    "VersionNotSupported",
    "Forbidden",
    "NotFound",
    "UnprocessableEntity",
    "NotImplementedResponse",
    "POSSIBLE_ERRORS",
)


class StrReprMixin(HTTPException):
    """This mixin can be useful when testing requires a string
    representation of an exception that contains the HTTPException
    detail string, rather than the standard Python exception message.
    """

    def __str__(self):
        return self.detail if self.detail is not None else self.__repr__()


class BadRequest(StrReprMixin, HTTPException):
    """400 Bad Request"""

    status_code: int = 400
    title: str = "Bad Request"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class VersionNotSupported(StrReprMixin, HTTPException):
    """553 Version Not Supported"""

    status_code: int = 553
    title: str = "Version Not Supported"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class Forbidden(StrReprMixin, HTTPException):
    """403 Forbidden"""

    status_code: int = 403
    title: str = "Forbidden"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class NotFound(StrReprMixin, HTTPException):
    """404 Not Found"""

    status_code: int = 404
    title: str = "Not Found"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class UnprocessableEntity(StrReprMixin, HTTPException):
    """422 Unprocessable Entity"""

    status_code: int = 422
    title: str = "Unprocessable Entity"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class InternalServerError(StrReprMixin, HTTPException):
    """500 Internal Server Error"""

    status_code: int = 500
    title: str = "Internal Server Error"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


class NotImplementedResponse(StrReprMixin, HTTPException):
    """501 Not Implemented"""

    status_code: int = 501
    title: str = "Not Implemented"

    def __init__(
        self,
        detail: str = None,
        headers: dict = None,
    ) -> None:
        super().__init__(status_code=self.status_code, detail=detail, headers=headers)


POSSIBLE_ERRORS = (
    BadRequest,
    Forbidden,
    NotFound,
    UnprocessableEntity,
    InternalServerError,
    NotImplementedResponse,
    VersionNotSupported,
)



"""This submodule implements the [`LarkParser`][optimade.filterparser.LarkParser] class,
which uses the lark library to parse filter strings with a defined OPTIMADE filter grammar
into `Lark.Tree` objects for use by the filter transformers.

"""

from pathlib import Path
from typing import Dict, Tuple
from collections import defaultdict

from lark import Lark, Tree

from optimade.server.exceptions import BadRequest

__all__ = ("ParserError", "LarkParser")


class ParserError(Exception):
    """Triggered by critical parsing errors that should lead
    to 500 Server Error HTTP statuses.
    """


def get_versions() -> Dict[Tuple[int, int, int], Dict[str, str]]:
    """Find grammar files within this package's grammar directory,
    returning a dictionary broken down by scraped grammar version
    (major, minor, patch) and variant (a string tag).

    Returns:
        A mapping from version, variant to grammar file name.

    """
    dct = defaultdict(dict)
    for filename in Path(__file__).parent.joinpath("../grammar").glob("*.lark"):
        tags = filename.stem.lstrip("v").split(".")
        version = tuple(map(int, tags[:3]))
        variant = "default" if len(tags) == 3 else tags[-1]
        dct[version][variant] = filename
    return dict(dct)


AVAILABLE_PARSERS = get_versions()


class LarkParser:
    """This class wraps a versioned OPTIMADE grammar and allows
    it to be parsed into Lark tree objects.

    """

    def __init__(self, version: Tuple[int, int, int] = None, variant: str = "default"):
        """For a given version and variant, try to load the corresponding grammar.

        Parameters:
            version: The grammar version number to use (e.g., `(1, 0, 1)` for v1.0.1).
            variant: The grammar variant to employ.

        Raises:
            ParserError: If the requested version/variant of the
                grammar does not exist.

        """

        version = version if version else max(AVAILABLE_PARSERS.keys())

        if version not in AVAILABLE_PARSERS:
            raise ParserError(f"Unknown parser grammar version: {version}")

        if variant not in AVAILABLE_PARSERS[version]:
            raise ParserError(f"Unknown variant of the parser: {variant}")

        self.version = version
        self.variant = variant

        with open(AVAILABLE_PARSERS[version][variant]) as f:
            self.lark = Lark(f)

        self.tree = None
        self.filter = None

    def parse(self, filter_: str) -> Tree:
        """Parse a filter string into a `lark.Tree`.

        Parameters:
            filter_: The filter string to parse.

        Raises:
            BadRequest: If the filter cannot be parsed.

        Returns:
            The parsed filter.

        """
        try:
            self.tree = self.lark.parse(filter_)
            self.filter = filter_
            return self.tree
        except Exception as exc:
            raise BadRequest(
                detail=f"Unable to parse filter {filter_}. Lark traceback: \n{exc}"
            ) from exc

    def __repr__(self):
        if isinstance(self.tree, Tree):
            return self.tree.pretty()
        return repr(self.lark)



# pylint: disable=line-too-long,no-self-argument
from datetime import datetime
from typing import Optional, Dict, List
from pydantic import BaseModel, Field, validator  # pylint: disable=no-name-in-module

from optimade.models.jsonapi import Relationships, Attributes, Resource
from optimade.models.optimade_json import Relationship, DataType
from optimade.models.utils import StrictField, OptimadeField, SupportLevel


__all__ = (
    "EntryRelationships",
    "EntryResourceAttributes",
    "EntryResource",
    "EntryInfoProperty",
    "EntryInfoResource",
)


class TypedRelationship(Relationship):
    # This may be updated when moving to Python 3.8
    @validator("data")
    def check_rel_type(cls, data):
        if not isinstance(data, list):
            # All relationships at this point are empty-to-many relationships in JSON:API:
            # https://jsonapi.org/format/1.0/#document-resource-object-linkage
            raise ValueError("`data` key in a relationship must always store a list.")
        if hasattr(cls, "_req_type") and any(
            getattr(obj, "type", None) != cls._req_type for obj in data
        ):
            raise ValueError("Object stored in relationship data has wrong type")
        return data


class ReferenceRelationship(TypedRelationship):
    _req_type = "references"


class StructureRelationship(TypedRelationship):
    _req_type = "structures"


class EntryRelationships(Relationships):
    """This model wraps the JSON API Relationships to include type-specific top level keys."""

    references: Optional[ReferenceRelationship] = StrictField(
        None,
        description="Object containing links to relationships with entries of the `references` type.",
    )

    structures: Optional[StructureRelationship] = StrictField(
        None,
        description="Object containing links to relationships with entries of the `structures` type.",
    )


class EntryResourceAttributes(Attributes):
    """Contains key-value pairs representing the entry's properties."""

    immutable_id: Optional[str] = OptimadeField(
        None,
        description="""The entry's immutable ID (e.g., an UUID). This is important for databases having preferred IDs that point to "the latest version" of a record, but still offer access to older variants. This ID maps to the version-specific record, in case it changes in the future.

- **Type**: string.

- **Requirements/Conventions**:
    - **Support**: OPTIONAL support in implementations, i.e., MAY be `null`.
    - **Query**: MUST be a queryable property with support for all mandatory filter features.

- **Examples**:
    - `"8bd3e750-b477-41a0-9b11-3a799f21b44f"`
    - `"fjeiwoj,54;@=%<>#32"` (Strings that are not URL-safe are allowed.)""",
        support=SupportLevel.OPTIONAL,
        queryable=SupportLevel.MUST,
    )

    last_modified: Optional[datetime] = OptimadeField(
        ...,
        description="""Date and time representing when the entry was last modified.

- **Type**: timestamp.

- **Requirements/Conventions**:
    - **Support**: SHOULD be supported by all implementations, i.e., SHOULD NOT be `null`.
    - **Query**: MUST be a queryable property with support for all mandatory filter features.
    - **Response**: REQUIRED in the response unless the query parameter `response_fields` is present and does not include this property.

- **Example**:
    - As part of JSON response format: `"2007-04-05T14:30:20Z"` (i.e., encoded as an [RFC 3339 Internet Date/Time Format](https://tools.ietf.org/html/rfc3339#section-5.6) string.)""",
        support=SupportLevel.SHOULD,
        queryable=SupportLevel.MUST,
    )


class EntryResource(Resource):
    """The base model for an entry resource."""

    id: str = OptimadeField(
        ...,
        description="""An entry's ID as defined in section Definition of Terms.

- **Type**: string.

- **Requirements/Conventions**:
    - **Support**: MUST be supported by all implementations, MUST NOT be `null`.
    - **Query**: MUST be a queryable property with support for all mandatory filter features.
    - **Response**: REQUIRED in the response.

- **Examples**:
    - `"db/1234567"`
    - `"cod/2000000"`
    - `"cod/2000000@1234567"`
    - `"nomad/L1234567890"`
    - `"42"`""",
        support=SupportLevel.MUST,
        queryable=SupportLevel.MUST,
    )

    type: str = Field(
        description="""The name of the type of an entry.

- **Type**: string.

- **Requirements/Conventions**:
    - **Support**: MUST be supported by all implementations, MUST NOT be `null`.
    - **Query**: MUST be a queryable property with support for all mandatory filter features.
    - **Response**: REQUIRED in the response.
    - MUST be an existing entry type.
    - The entry of type `<type>` and ID `<id>` MUST be returned in response to a request for `/<type>/<id>` under the versioned base URL.

- **Example**: `"structures"`""",
        support=SupportLevel.MUST,
        queryable=SupportLevel.MUST,
    )

    attributes: EntryResourceAttributes = StrictField(
        ...,
        description="""A dictionary, containing key-value pairs representing the entry's properties, except for `type` and `id`.
Database-provider-specific properties need to include the database-provider-specific prefix (see section on Database-Provider-Specific Namespace Prefixes).""",
    )

    relationships: Optional[EntryRelationships] = StrictField(
        None,
        description="""A dictionary containing references to other entries according to the description in section Relationships encoded as [JSON API Relationships](https://jsonapi.org/format/1.0/#document-resource-object-relationships).
The OPTIONAL human-readable description of the relationship MAY be provided in the `description` field inside the `meta` dictionary of the JSON API resource identifier object.""",
    )


class EntryInfoProperty(BaseModel):

    description: str = StrictField(
        ..., description="A human-readable description of the entry property"
    )

    unit: Optional[str] = StrictField(
        None,
        description="""The physical unit of the entry property.
This MUST be a valid representation of units according to version 2.1 of [The Unified Code for Units of Measure](https://unitsofmeasure.org/ucum.html).
It is RECOMMENDED that non-standard (non-SI) units are described in the description for the property.""",
    )

    sortable: Optional[bool] = StrictField(
        None,
        description="""Defines whether the entry property can be used for sorting with the "sort" parameter.
If the entry listing endpoint supports sorting, this key MUST be present for sortable properties with value `true`.""",
    )

    type: Optional[DataType] = StrictField(
        None,
        title="Type",
        description="""The type of the property's value.
This MUST be any of the types defined in the Data types section.
For the purpose of compatibility with future versions of this specification, a client MUST accept values that are not `string` values specifying any of the OPTIMADE Data types, but MUST then also disregard the `type` field.
Note, if the value is a nested type, only the outermost type should be reported.
E.g., for the entry resource `structures`, the `species` property is defined as a list of dictionaries, hence its `type` value would be `list`.""",
    )


class EntryInfoResource(BaseModel):

    formats: List[str] = StrictField(
        ..., description="List of output formats available for this type of entry."
    )

    description: str = StrictField(..., description="Description of the entry.")

    properties: Dict[str, EntryInfoProperty] = StrictField(
        ...,
        description="A dictionary describing queryable properties for this entry type, where each key is a property name.",
    )

    output_fields_by_format: Dict[str, List[str]] = StrictField(
        ...,
        description="Dictionary of available output fields for this entry type, where the keys are the values of the `formats` list and the values are the keys of the `properties` dictionary.",
    )



from fastapi import Query
from pydantic import EmailStr  # pylint: disable=no-name-in-module

from optimade.server.config import CONFIG


class EntryListingQueryParams:
    """
    Common query params for all Entry listing endpoints.

    Attributes:
        filter (str): A filter string, in the format described in section API Filtering Format Specification of the specification.

        response_format (str): The output format requested (see section Response Format).
            Defaults to the format string 'json', which specifies the standard output format described in this specification.

            **Example**: `http://example.com/v1/structures?response_format=xml`

        email_address (EmailStr): An email address of the user making the request.
            The email SHOULD be that of a person and not an automatic system.

            **Example**: `http://example.com/v1/structures?email_address=user@example.com`

        response_fields (str): A comma-delimited set of fields to be provided in the output.
            If provided, these fields MUST be returned along with the REQUIRED fields.
            Other OPTIONAL fields MUST NOT be returned when this parameter is present.

            **Example**: `http://example.com/v1/structures?response_fields=last_modified,nsites`

        sort (str): If supporting sortable queries, an implementation MUST use the `sort` query parameter with format as specified
            by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-sorting).

            An implementation MAY support multiple sort fields for a single query.
            If it does, it again MUST conform to the JSON API 1.0 specification.

            If an implementation supports sorting for an entry listing endpoint, then the `/info/<entries>` endpoint MUST include,
            for each field name `<fieldname>` in its `data.properties.<fieldname>` response value that can be used for sorting,
            the key `sortable` with value `true`.
            If a field name under an entry listing endpoint supporting sorting cannot be used for sorting, the server MUST either
            leave out the `sortable` key or set it equal to `false` for the specific field name.
            The set of field names, with `sortable` equal to `true` are allowed to be used in the "sort fields" list according to
            its definition in the JSON API 1.0 specification.
            The field `sortable` is in addition to each property description and other OPTIONAL fields.
            An example is shown in the section Entry Listing Info Endpoints.

        page_limit (int): Sets a numerical limit on the number of entries returned.
            See [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-pagination).
            The API implementation MUST return no more than the number specified.
            It MAY return fewer. The database MAY have a maximum limit and not accept larger numbers
            (in which case an error code -- `403 Forbidden` -- MUST be returned).
            The default limit value is up to the API implementation to decide.

            **Example**: `http://example.com/optimade/v1/structures?page_limit=100`

        page_offset (int): RECOMMENDED for use with _offset-based_ pagination: using `page_offset` and `page_limit` is RECOMMENDED.

            **Example**: Skip 50 structures and fetch up to 100: `/structures?page_offset=50&page_limit=100`.

        page_number (int): RECOMMENDED for use with _page-based_ pagination: using `page_number` and `page_limit` is RECOMMENDED.
            It is RECOMMENDED that the first page has number 1, i.e., that `page_number` is 1-based.

            **Example**: Fetch page 2 of up to 50 structures per page: `/structures?page_number=2&page_limit=50`.

        page_cursor (int): RECOMMENDED for use with _cursor-based_ pagination: using `page_cursor` and `page_limit` is RECOMMENDED.

        page_above (int): RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.

            **Example**: Fetch up to 100 structures above sort-field value 4000 (in this example, server chooses to fetch results sorted by
            increasing `id`, so `page_above` value refers to an `id` value): `/structures?page_above=4000&page_limit=100`.

        page_below (int): RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.

        include (str): A server MAY implement the JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents)
            by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).

            All related resource objects MUST be returned as part of an array value for the top-level `included` field,
            see the section JSON Response Schema: Common Fields.

            The value of `include` MUST be a comma-separated list of "relationship paths", as defined in the [JSON API](https://jsonapi.org/format/1.0/#fetching-includes).
            If relationship paths are not supported, or a server is unable to identify a relationship path a `400 Bad Request` response MUST be made.

            The **default value** for `include` is `references`. This means `references` entries MUST always be included under the top-level field
            `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified as `include=references`.
            Note, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects MUST NOT be included under the top-level
            field `included`, as per the definition of `included`, see section JSON Response Schema: Common Fields.

            **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the top-level field `included`.

        api_hint (str): If the client provides the parameter, the value SHOULD have the format `vMAJOR` or `vMAJOR.MINOR`,
            where MAJOR is a major version and MINOR is a minor version of the API.
            For example, if a client appends `api_hint=v1.0` to the query string, the hint provided is for major version 1 and minor version 0.

    """

    def __init__(
        self,
        *,
        filter: str = Query(  # pylint: disable=redefined-builtin
            "",
            description="A filter string, in the format described in section API Filtering Format Specification of the specification.",
        ),
        response_format: str = Query(
            "json",
            description="The output format requested (see section Response Format).\nDefaults to the format string 'json', which specifies the standard output format described in this specification.\nExample: `http://example.com/v1/structures?response_format=xml`",
        ),
        email_address: EmailStr = Query(
            "",
            description="An email address of the user making the request.\nThe email SHOULD be that of a person and not an automatic system.\nExample: `http://example.com/v1/structures?email_address=user@example.com`",
        ),
        response_fields: str = Query(
            "",
            description="A comma-delimited set of fields to be provided in the output.\nIf provided, these fields MUST be returned along with the REQUIRED fields.\nOther OPTIONAL fields MUST NOT be returned when this parameter is present.\nExample: `http://example.com/v1/structures?response_fields=last_modified,nsites`",
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        sort: str = Query(
            "",
            description='If supporting sortable queries, an implementation MUST use the `sort` query parameter with format as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-sorting).\n\nAn implementation MAY support multiple sort fields for a single query.\nIf it does, it again MUST conform to the JSON API 1.0 specification.\n\nIf an implementation supports sorting for an entry listing endpoint, then the `/info/<entries>` endpoint MUST include, for each field name `<fieldname>` in its `data.properties.<fieldname>` response value that can be used for sorting, the key `sortable` with value `true`.\nIf a field name under an entry listing endpoint supporting sorting cannot be used for sorting, the server MUST either leave out the `sortable` key or set it equal to `false` for the specific field name.\nThe set of field names, with `sortable` equal to `true` are allowed to be used in the "sort fields" list according to its definition in the JSON API 1.0 specification.\nThe field `sortable` is in addition to each property description and other OPTIONAL fields.\nAn example is shown in the section Entry Listing Info Endpoints.',
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        page_limit: int = Query(
            CONFIG.page_limit,
            description="Sets a numerical limit on the number of entries returned.\nSee [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-pagination).\nThe API implementation MUST return no more than the number specified.\nIt MAY return fewer.\nThe database MAY have a maximum limit and not accept larger numbers (in which case an error code -- 403 Forbidden -- MUST be returned).\nThe default limit value is up to the API implementation to decide.\nExample: `http://example.com/optimade/v1/structures?page_limit=100`",
            ge=0,
        ),
        page_offset: int = Query(
            0,
            description="RECOMMENDED for use with _offset-based_ pagination: using `page_offset` and `page_limit` is RECOMMENDED.\nExample: Skip 50 structures and fetch up to 100: `/structures?page_offset=50&page_limit=100`.",
            ge=0,
        ),
        page_number: int = Query(
            0,
            description="RECOMMENDED for use with _page-based_ pagination: using `page_number` and `page_limit` is RECOMMENDED.\nIt is RECOMMENDED that the first page has number 1, i.e., that `page_number` is 1-based.\nExample: Fetch page 2 of up to 50 structures per page: `/structures?page_number=2&page_limit=50`.",
            ge=0,
        ),
        page_cursor: int = Query(
            0,
            description="RECOMMENDED for use with _cursor-based_ pagination: using `page_cursor` and `page_limit` is RECOMMENDED.",
            ge=0,
        ),
        page_above: int = Query(
            0,
            description="RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.\nExample: Fetch up to 100 structures above sort-field value 4000 (in this example, server chooses to fetch results sorted by increasing `id`, so `page_above` value refers to an `id` value): `/structures?page_above=4000&page_limit=100`.",
            ge=0,
        ),
        page_below: int = Query(
            0,
            description="RECOMMENDED for use with _value-based_ pagination: using `page_above`/`page_below` and `page_limit` is RECOMMENDED.",
            ge=0,
        ),
        include: str = Query(
            "references",
            description='A server MAY implement the JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents) by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).\n\nAll related resource objects MUST be returned as part of an array value for the top-level `included` field, see the section JSON Response Schema: Common Fields.\n\nThe value of `include` MUST be a comma-separated list of "relationship paths", as defined in the [JSON API](https://jsonapi.org/format/1.0/#fetching-includes).\nIf relationship paths are not supported, or a server is unable to identify a relationship path a `400 Bad Request` response MUST be made.\n\nThe **default value** for `include` is `references`.\nThis means `references` entries MUST always be included under the top-level field `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified as `include=references`.\nNote, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects MUST NOT be included under the top-level field `included`, as per the definition of `included`, see section JSON Response Schema: Common Fields.\n\n> **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the top-level field `included`.',
        ),
        api_hint: str = Query(
            "",
            description="If the client provides the parameter, the value SHOULD have the format `vMAJOR` or `vMAJOR.MINOR`, where MAJOR is a major version and MINOR is a minor version of the API. For example, if a client appends `api_hint=v1.0` to the query string, the hint provided is for major version 1 and minor version 0.",
            regex=r"(v[0-9]+(\.[0-9]+)?)?",
        ),
    ):
        self.filter = filter
        self.response_format = response_format
        self.email_address = email_address
        self.response_fields = response_fields
        self.sort = sort
        self.page_limit = page_limit
        self.page_offset = page_offset
        self.page_number = page_number
        self.page_cursor = page_cursor
        self.page_above = page_above
        self.page_below = page_below
        self.include = include


class SingleEntryQueryParams:
    """
    Common query params for single entry endpoints.

    Attributes:
        response_format (str): The output format requested (see section Response Format).
            Defaults to the format string 'json', which specifies the standard output format described in this specification.

            **Example**: `http://example.com/v1/structures?response_format=xml`

        email_address (EmailStr): An email address of the user making the request.
            The email SHOULD be that of a person and not an automatic system.

            **Example**: `http://example.com/v1/structures?email_address=user@example.com`

        response_fields (str): A comma-delimited set of fields to be provided in the output.
            If provided, these fields MUST be returned along with the REQUIRED fields.
            Other OPTIONAL fields MUST NOT be returned when this parameter is present.

            **Example**: `http://example.com/v1/structures?response_fields=last_modified,nsites`

        include (str): A server MAY implement the JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents)
            by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).

            All related resource objects MUST be returned as part of an array value for the top-level `included` field,
            see the section JSON Response Schema: Common Fields.

            The value of `include` MUST be a comma-separated list of "relationship paths", as defined in the [JSON API](https://jsonapi.org/format/1.0/#fetching-includes).
            If relationship paths are not supported, or a server is unable to identify a relationship path a `400 Bad Request` response MUST be made.

            The **default value** for `include` is `references`. This means `references` entries MUST always be included under the top-level field
            `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified as `include=references`.
            Note, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects MUST NOT be included under the top-level
            field `included`, as per the definition of `included`, see section JSON Response Schema: Common Fields.

            **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the top-level field `included`.

        api_hint (str): If the client provides the parameter, the value SHOULD have the format `vMAJOR` or `vMAJOR.MINOR`,
            where MAJOR is a major version and MINOR is a minor version of the API.
            For example, if a client appends `api_hint=v1.0` to the query string, the hint provided is for major version 1 and minor version 0.

    """

    def __init__(
        self,
        *,
        response_format: str = Query(
            "json",
            description="The output format requested (see section Response Format).\nDefaults to the format string 'json', which specifies the standard output format described in this specification.\nExample: `http://example.com/v1/structures?response_format=xml`",
        ),
        email_address: EmailStr = Query(
            "",
            description="An email address of the user making the request.\nThe email SHOULD be that of a person and not an automatic system.\nExample: `http://example.com/v1/structures?email_address=user@example.com`",
        ),
        response_fields: str = Query(
            "",
            description="A comma-delimited set of fields to be provided in the output.\nIf provided, these fields MUST be returned along with the REQUIRED fields.\nOther OPTIONAL fields MUST NOT be returned when this parameter is present.\nExample: `http://example.com/v1/structures?response_fields=last_modified,nsites`",
            regex=r"([a-z_][a-z_0-9]*(,[a-z_][a-z_0-9]*)*)?",
        ),
        include: str = Query(
            "references",
            description='A server MAY implement the JSON API concept of returning [compound documents](https://jsonapi.org/format/1.0/#document-compound-documents) by utilizing the `include` query parameter as specified by [JSON API 1.0](https://jsonapi.org/format/1.0/#fetching-includes).\n\nAll related resource objects MUST be returned as part of an array value for the top-level `included` field, see the section JSON Response Schema: Common Fields.\n\nThe value of `include` MUST be a comma-separated list of "relationship paths", as defined in the [JSON API](https://jsonapi.org/format/1.0/#fetching-includes).\nIf relationship paths are not supported, or a server is unable to identify a relationship path a `400 Bad Request` response MUST be made.\n\nThe **default value** for `include` is `references`.\nThis means `references` entries MUST always be included under the top-level field `included` as default, since a server assumes if `include` is not specified by a client in the request, it is still specified as `include=references`.\nNote, if a client explicitly specifies `include` and leaves out `references`, `references` resource objects MUST NOT be included under the top-level field `included`, as per the definition of `included`, see section JSON Response Schema: Common Fields.\n\n> **Note**: A query with the parameter `include` set to the empty string means no related resource objects are to be returned under the top-level field `included`.',
        ),
        api_hint: str = Query(
            "",
            description="If the client provides the parameter, the value SHOULD have the format `vMAJOR` or `vMAJOR.MINOR`, where MAJOR is a major version and MINOR is a minor version of the API. For example, if a client appends `api_hint=v1.0` to the query string, the hint provided is for major version 1 and minor version 0.",
            regex=r"(v[0-9]+(\.[0-9]+)?)?",
        ),
    ):
        self.response_format = response_format
        self.email_address = email_address
        self.response_fields = response_fields
        self.include = include



from fastapi import APIRouter, Request

from optimade import __api_version__
from optimade.models import (
    IndexInfoResponse,
    IndexInfoAttributes,
    IndexInfoResource,
    IndexRelationship,
    RelatedLinksResource,
)
from optimade.server.config import CONFIG
from optimade.server.routers.utils import meta_values, get_base_url
from optimade.server.schemas import ERROR_RESPONSES


router = APIRouter(redirect_slashes=True)


@router.get(
    "/info",
    response_model=IndexInfoResponse,
    response_model_exclude_unset=True,
    tags=["Info"],
    responses=ERROR_RESPONSES,
)
def get_info(request: Request) -> IndexInfoResponse:
    return IndexInfoResponse(
        meta=meta_values(request.url, 1, 1, more_data_available=False),
        data=IndexInfoResource(
            id=IndexInfoResource.schema()["properties"]["id"]["const"],
            type=IndexInfoResource.schema()["properties"]["type"]["const"],
            attributes=IndexInfoAttributes(
                api_version=f"{__api_version__}",
                available_api_versions=[
                    {
                        "url": f"{get_base_url(request.url)}/v{__api_version__.split('.')[0]}/",
                        "version": f"{__api_version__}",
                    }
                ],
                formats=["json"],
                available_endpoints=["info", "links"],
                entry_types_by_format={"json": []},
                is_index=True,
            ),
            relationships={
                "default": IndexRelationship(
                    data={
                        "type": RelatedLinksResource.schema()["properties"]["type"][
                            "const"
                        ],
                        "id": CONFIG.default_db,
                    }
                )
            },
        ),
    )



# pylint: disable=no-self-argument
from enum import Enum

from pydantic import Field, BaseModel  # pylint: disable=no-name-in-module
from typing import Union, Dict

from optimade.models.jsonapi import BaseResource
from optimade.models.baseinfo import BaseInfoAttributes, BaseInfoResource
from optimade.models.utils import StrictField


__all__ = (
    "IndexInfoAttributes",
    "RelatedLinksResource",
    "IndexRelationship",
    "IndexInfoResource",
)


class DefaultRelationship(Enum):
    """Enumeration of key(s) for relationship dictionary in IndexInfoResource"""

    DEFAULT = "default"


class IndexInfoAttributes(BaseInfoAttributes):
    """Attributes for Base URL Info endpoint for an Index Meta-Database"""

    is_index: bool = StrictField(
        True,
        const=True,
        description="This must be `true` since this is an index meta-database (see section Index Meta-Database).",
    )


class RelatedLinksResource(BaseResource):
    """A related Links resource object"""

    type: str = Field("links", const="links", regex="^links$")


class IndexRelationship(BaseModel):
    """Index Meta-Database relationship"""

    data: Union[None, RelatedLinksResource] = StrictField(
        ...,
        description="""[JSON API resource linkage](http://jsonapi.org/format/1.0/#document-links).
It MUST be either `null` or contain a single Links identifier object with the fields `id` and `type`""",
    )


class IndexInfoResource(BaseInfoResource):
    """Index Meta-Database Base URL Info endpoint resource"""

    attributes: IndexInfoAttributes = Field(...)
    relationships: Union[
        None, Dict[DefaultRelationship, IndexRelationship]
    ] = StrictField(
        ...,
        title="Relationships",
        description="""Reference to the Links identifier object under the `links` endpoint that the provider has chosen as their 'default' OPTIMADE API database.
A client SHOULD present this database as the first choice when an end-user chooses this provider.""",
    )



from typing import Dict, Tuple, List, Any

from optimade.filterparser import LarkParser
from optimade.filtertransformers.mongo import MongoTransformer
from optimade.models import EntryResource
from optimade.server.config import CONFIG
from optimade.server.entry_collections import EntryCollection
from optimade.server.logger import LOGGER
from optimade.server.mappers import BaseResourceMapper


if CONFIG.database_backend.value == "mongodb":
    from pymongo import MongoClient

    LOGGER.info("Using: Real MongoDB (pymongo)")

elif CONFIG.database_backend.value == "mongomock":
    from mongomock import MongoClient

    LOGGER.info("Using: Mock MongoDB (mongomock)")

if CONFIG.database_backend.value in ("mongomock", "mongodb"):
    CLIENT = MongoClient(CONFIG.mongo_uri)


class MongoCollection(EntryCollection):
    """Class for querying MongoDB collections (implemented by either pymongo or mongomock)
    containing serialized [`EntryResource`][optimade.models.entries.EntryResource]s objects.

    """

    def __init__(
        self,
        name: str,
        resource_cls: EntryResource,
        resource_mapper: BaseResourceMapper,
        database: str = CONFIG.mongo_database,
    ):
        """Initialize the MongoCollection for the given parameters.

        Parameters:
            name: The name of the collection.
            resource_cls: The type of entry resource that is stored by the collection.
            resource_mapper: A resource mapper object that handles aliases and
                format changes between deserialization and response.
            database: The name of the underlying MongoDB database to connect to.

        """
        super().__init__(
            resource_cls,
            resource_mapper,
            MongoTransformer(mapper=resource_mapper),
        )

        self.parser = LarkParser(version=(1, 0, 0), variant="default")
        self.collection = CLIENT[database][name]

        # check aliases do not clash with mongo operators
        self._check_aliases(self.resource_mapper.all_aliases())
        self._check_aliases(self.resource_mapper.all_length_aliases())

    def __len__(self) -> int:
        """Returns the total number of entries in the collection."""
        return self.collection.estimated_document_count()

    def count(self, **kwargs) -> int:
        """Returns the number of entries matching the query specified
        by the keyword arguments.

        Parameters:
            kwargs (dict): Query parameters as keyword arguments. The keys
                'filter', 'skip', 'limit', 'hint' and 'maxTimeMS' will be passed
                to the `pymongo.collection.Collection.count_documents` method.

        """
        for k in list(kwargs.keys()):
            if k not in ("filter", "skip", "limit", "hint", "maxTimeMS"):
                del kwargs[k]
        if "filter" not in kwargs:  # "filter" is needed for count_documents()
            kwargs["filter"] = {}
        return self.collection.count_documents(**kwargs)

    def insert(self, data: List[EntryResource]) -> None:
        """Add the given entries to the underlying database.

        Warning:
            No validation is performed on the incoming data.

        Arguments:
            data: The entry resource objects to add to the database.

        """
        self.collection.insert_many(data)

    def _run_db_query(
        self, criteria: Dict[str, Any], single_entry: bool = False
    ) -> Tuple[List[Dict[str, Any]], int, bool]:
        """Run the query on the backend and collect the results.

        Arguments:
            criteria: A dictionary representation of the query parameters.
            single_entry: Whether or not the caller is expecting a single entry response.

        Returns:
            The list of entries from the database (without any re-mapping), the total number of
            entries matching the query and a boolean for whether or not there is more data available.

        """

        results = []
        for doc in self.collection.find(**criteria):
            if criteria.get("projection", {}).get("_id"):
                doc["_id"] = str(doc["_id"])
            results.append(doc)

        nresults_now = len(results)
        if not single_entry:
            criteria_nolimit = criteria.copy()
            criteria_nolimit.pop("limit", None)
            data_returned = self.count(**criteria_nolimit)
            more_data_available = nresults_now < data_returned
        else:
            # SingleEntryQueryParams, e.g., /structures/{entry_id}
            data_returned = nresults_now
            more_data_available = False

        return results, data_returned, more_data_available

    def _check_aliases(self, aliases):
        """Check that aliases do not clash with mongo keywords."""
        if any(
            alias[0].startswith("$") or alias[1].startswith("$") for alias in aliases
        ):
            raise RuntimeError(f"Cannot define an alias starting with a '$': {aliases}")



import inspect
import warnings
import re
import itertools
from enum import Enum
from typing import Optional

from pydantic import Field

_PYDANTIC_FIELD_KWARGS = list(inspect.signature(Field).parameters.keys())

__all__ = (
    "CHEMICAL_SYMBOLS",
    "EXTRA_SYMBOLS",
    "ATOMIC_NUMBERS",
    "SemanticVersion",
    "SupportLevel",
)


class SupportLevel(Enum):
    """OPTIMADE property/field support levels"""

    MUST = "must"
    SHOULD = "should"
    OPTIONAL = "optional"


def StrictField(
    *args,
    description: str = None,
    **kwargs,
) -> Field:
    """A wrapper around `pydantic.Field` that does the following:

    - Forbids any "extra" keys that would be passed to `pydantic.Field`,
      except those used elsewhere to modify the schema in-place,
      e.g. "uniqueItems", "pattern" and those added by OptimadeField, e.g.
      "unit", "queryable" and "sortable".
    - Emits a warning when no description is provided.

    Arguments:
        *args: Positional arguments passed through to `Field`.
        description: The description of the `Field`; if this is not
            specified then a `UserWarning` will be emitted.
        **kwargs: Extra keyword arguments to be passed to `Field`.

    Raises:
        RuntimeError: If `**kwargs` contains a key not found in the
            function signature of `Field`, or in the extensions used
            by models in this package (see above).

    Returns:
        The pydantic `Field`.

    """

    allowed_keys = [
        "unit",
        "pattern",
        "uniqueItems",
        "support",
        "queryable",
        "sortable",
    ]
    _banned = [k for k in kwargs if k not in set(_PYDANTIC_FIELD_KWARGS + allowed_keys)]

    if _banned:
        raise RuntimeError(
            f"Not creating StrictField({args}, {kwargs}) with forbidden keywords {_banned}."
        )

    if description is not None:
        kwargs["description"] = description

    if description is None:
        warnings.warn(
            f"No description provided for StrictField specified by {args}, {kwargs}."
        )

    return Field(*args, **kwargs)


def OptimadeField(
    *args,
    support: Optional[SupportLevel] = None,
    queryable: Optional[SupportLevel] = None,
    unit: Optional[str] = None,
    **kwargs,
) -> Field:
    """A wrapper around `pydantic.Field` that adds OPTIMADE-specific
    field paramters `queryable`, `support` and `unit`, indicating
    the corresponding support level in the specification and the
    physical unit of the field.

    Arguments:
        support: The support level of the field itself, i.e. whether the field
            can be null or omitted by an implementation.
        queryable: The support level corresponding to the queryablility
            of this field.
        unit: A string describing the unit of the field.

    Returns:
        The pydantic field with extra validation provided by [`StrictField`][optimade.models.utils.StrictField].

    """

    # Collect non-null keyword arguments to add to the Field schema
    if unit is not None:
        kwargs["unit"] = unit
    if queryable is not None:
        if isinstance(queryable, str):
            queryable = SupportLevel(queryable.lower())
        kwargs["queryable"] = queryable
    if support is not None:
        if isinstance(support, str):
            support = SupportLevel(support.lower())
        kwargs["support"] = support

    return StrictField(*args, **kwargs)


class SemanticVersion(str):
    """A custom type for a semantic version, using the recommended
    semver regexp from
    https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string.

    """

    regex = re.compile(
        r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
    )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
            pattern=cls.regex.pattern,
            examples=["0.10.1", "1.0.0-rc.2", "1.2.3-rc.5+develop"],
        )

    @classmethod
    def validate(cls, v: str):
        if not cls.regex.match(v):
            raise ValueError(
                f"Unable to validate the version string {v!r} as a semantic version (expected <major>.<minor>.<patch>)."
                "See https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string for more information."
            )

        return v

    @property
    def _match(self):
        """The result of the regex match."""
        return self.regex.match(self)

    @property
    def major(self) -> int:
        """The major version number."""
        return int(self._match.group(1))

    @property
    def minor(self) -> int:
        """The minor version number."""
        return int(self._match.group(2))

    @property
    def patch(self) -> int:
        """The patch version number."""
        return int(self._match.group(3))

    @property
    def prerelease(self) -> str:
        """The pre-release tag."""
        return self._match.group(4)

    @property
    def build_metadata(self) -> str:
        """The build metadata."""
        return self._match.group(5)

    @property
    def base_version(self) -> str:
        """The base version string without patch and metadata info."""
        return f"{self.major}.{self.minor}.{self.patch}"


def anonymous_element_generator():
    """Generator that yields the next symbol in the A, B, Aa, ... Az naming scheme."""
    from string import ascii_lowercase

    for size in itertools.count(1):
        for s in itertools.product(ascii_lowercase, repeat=size):
            s = list(s)
            s[0] = s[0].upper()
            yield "".join(s)


ANONYMOUS_ELEMENTS = tuple(itertools.islice(anonymous_element_generator(), 150))
""" Returns the first 150 values of the anonymous element generator. """

CHEMICAL_FORMULA_REGEXP = r"^([A-Z][a-z]?\d*)*$"

EXTRA_SYMBOLS = ["X", "vacancy"]

CHEMICAL_SYMBOLS = [
    "H",
    "He",
    "Li",
    "Be",
    "B",
    "C",
    "N",
    "O",
    "F",
    "Ne",
    "Na",
    "Mg",
    "Al",
    "Si",
    "P",
    "S",
    "Cl",
    "Ar",
    "K",
    "Ca",
    "Sc",
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Zn",
    "Ga",
    "Ge",
    "As",
    "Se",
    "Br",
    "Kr",
    "Rb",
    "Sr",
    "Y",
    "Zr",
    "Nb",
    "Mo",
    "Tc",
    "Ru",
    "Rh",
    "Pd",
    "Ag",
    "Cd",
    "In",
    "Sn",
    "Sb",
    "Te",
    "I",
    "Xe",
    "Cs",
    "Ba",
    "La",
    "Ce",
    "Pr",
    "Nd",
    "Pm",
    "Sm",
    "Eu",
    "Gd",
    "Tb",
    "Dy",
    "Ho",
    "Er",
    "Tm",
    "Yb",
    "Lu",
    "Hf",
    "Ta",
    "W",
    "Re",
    "Os",
    "Ir",
    "Pt",
    "Au",
    "Hg",
    "Tl",
    "Pb",
    "Bi",
    "Po",
    "At",
    "Rn",
    "Fr",
    "Ra",
    "Ac",
    "Th",
    "Pa",
    "U",
    "Np",
    "Pu",
    "Am",
    "Cm",
    "Bk",
    "Cf",
    "Es",
    "Fm",
    "Md",
    "No",
    "Lr",
    "Rf",
    "Db",
    "Sg",
    "Bh",
    "Hs",
    "Mt",
    "Ds",
    "Rg",
    "Cn",
    "Nh",
    "Fl",
    "Mc",
    "Lv",
    "Ts",
    "Og",
]

ATOMIC_NUMBERS = {}
for Z, symbol in enumerate(CHEMICAL_SYMBOLS):
    ATOMIC_NUMBERS[symbol] = Z + 1



from optimade.models import ReferenceResource
from optimade.adapters.base import EntryAdapter


class Reference(EntryAdapter):
    """
    Lazy reference resource converter.

    Go to [`EntryAdapter`][optimade.adapters.base.EntryAdapter] to see the full list of methods
    and properties.

    Attributes:
        ENTRY_RESOURCE (ReferenceResource): This adapter stores entry resources as
            [`ReferenceResource`][optimade.models.references.ReferenceResource]s.
        _type_converters (Dict[str, Callable]): Dictionary of valid conversion types for entry.

            There are currently no available types.
        as_<_type_converters>: Convert entry to a type listed in `_type_converters`.

    """

    ENTRY_RESOURCE: ReferenceResource = ReferenceResource



from pathlib import Path
import re
from setuptools import setup, find_packages

module_dir = Path(__file__).resolve().parent

with open(module_dir.joinpath("optimade/__init__.py")) as version_file:
    for line in version_file:
        match = re.match(r'__version__ = "(.*)"', line)
        if match is not None:
            VERSION = match.group(1)
            break
    else:
        raise RuntimeError(
            f"Could not determine package version from {version_file.name} !"
        )

# Dependencies
# Server minded
elastic_deps = ["elasticsearch-dsl~=6.4,<7.0"]
mongo_deps = ["pymongo~=3.11", "mongomock~=3.23"]
server_deps = [
    "uvicorn~=0.13.4",
    "Jinja2>=2.10,<4.0",
    "pyyaml~=5.1",  # Keep at pyyaml 5.1 for aiida-core support
] + mongo_deps

# Client minded
aiida_deps = ["aiida-core~=1.6"]
ase_deps = ["ase~=3.21"]
cif_deps = ["numpy~=1.20"]
pdb_deps = cif_deps
pymatgen_deps = ["pymatgen==2021.3.9"]
jarvis_deps = ["jarvis-tools==2021.5.28"]
client_deps = cif_deps

# General
docs_deps = [
    "mkdocs~=1.1",
    "mkdocs-awesome-pages-plugin~=2.5",
    "mkdocs-material~=7.1",
    "mkdocs-minify-plugin~=0.4.0",
    "mkdocstrings~=0.15.1",
]
testing_deps = [
    "pytest~=6.2",
    "pytest-cov~=2.12",
    "codecov~=2.1",
    "jsondiff~=1.3",
] + server_deps
dev_deps = (
    ["pylint~=2.8", "pre-commit~=2.13", "invoke~=1.5"]
    + docs_deps
    + testing_deps
    + client_deps
)
all_deps = dev_deps + elastic_deps + aiida_deps + ase_deps + pymatgen_deps + jarvis_deps

setup(
    name="optimade",
    version=VERSION,
    url="https://github.com/Materials-Consortia/optimade-python-tools",
    license="MIT",
    author="OPTIMADE Development Team",
    author_email="dev@optimade.org",
    description="Tools for implementing and consuming OPTIMADE APIs.",
    long_description=open(module_dir.joinpath("README.md")).read(),
    long_description_content_type="text/markdown",
    keywords="optimade jsonapi materials",
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Database :: Front-Ends",
    ],
    python_requires=">=3.7",
    install_requires=[
        "lark-parser~=0.11.3",
        "fastapi~=0.65.2",
        "pydantic~=1.8,>=1.8.2",
        "email_validator~=1.1",
        "requests~=2.25",
    ],
    extras_require={
        "all": all_deps,
        "dev": dev_deps,
        "docs": docs_deps,
        "testing": testing_deps,
        "server": server_deps,
        "client": client_deps,
        "elastic": elastic_deps,
        "mongo": mongo_deps,
        "aiida": aiida_deps,
        "ase": ase_deps,
        "cif": cif_deps,
        "pdb": pdb_deps,
        "pymatgen": pymatgen_deps,
        "jarvis": jarvis_deps,
    },
    entry_points={
        "console_scripts": ["optimade-validator=optimade.validator:validate"]
    },
)



from fastapi import Request, APIRouter
from fastapi.responses import Response

from optimade.server.routers.utils import BASE_URL_PREFIXES

router = APIRouter(redirect_slashes=True)


class CsvResponse(Response):
    media_type = "text/csv; header=present"


@router.get(
    "/versions",
    tags=["Versions"],
    response_class=CsvResponse,
)
def get_versions(request: Request) -> CsvResponse:
    """Respond with the text/csv representation for the served versions."""
    version = BASE_URL_PREFIXES["major"].replace("/v", "")
    response = f"version\n{version}"
    return CsvResponse(content=response)
